# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import time
import numpy
from nplinker.genomics import GCF
from nplinker.metabolomics import MolecularFamily
from . import iokr_opt
from . import iokrdata as iokrdataserver
from . import mk_fprints as fingerprint
from . import spectrum
from . import spectrum_filters


logger = logging.getLogger(__name__)


class NPLinkerIOKR:
    """Rank the potential links in a NPLinker object using IOKR, and wrap the results."""

    def __init__(self, npl):
        self.npl = npl

        # Build BGC<->SMILES lookup
        bgc_smiles_lookup = {}
        for bgc_idx, bgc in enumerate(npl.bgcs):
            bgc_smiles_lookup[bgc] = []
            if bgc.smiles is not None:
                try:
                    fingerprint.fingerprint_from_smiles(bgc.smiles)
                    bgc_smiles_lookup[bgc].append(bgc.smiles)
                except ValueError:
                    print(f"Filtered out smiles {bgc.smiles}")
        self.bgc_smiles_lookup = bgc_smiles_lookup

        smiles_bgc_lookup = {}
        for bgc, smiles in bgc_smiles_lookup.items():
            if smiles is None:
                continue
            for smiles_string in smiles:
                if smiles_string in smiles_bgc_lookup:
                    smiles_bgc_lookup[smiles_string].add(bgc)
                else:
                    smiles_bgc_lookup[smiles_string] = {bgc}

        self.smiles_list = list(smiles_bgc_lookup.keys())

        # Map NPLinker spectra to format understood by IOKR
        self.spectra_list = [spectrum.MSSpectrum(spec=x) for x in npl.spectra]

        # Initialise IOKR server
        self.iokr_server = get_iokr_server()

    def calculate_scores(self):
        self.iokr_scores = self.iokr_server.score_smiles(self.spectra_list, self.smiles_list)

    def get_score(self, mf_like, gcf_like):
        if isinstance(mf_like, MolecularFamily):
            mf = mf_like
        else:
            mf = MolecularFamily(-1)
            mf.spectra = [mf_like]

        if isinstance(gcf_like, GCF):
            gcf = gcf_like
        else:
            gcf = GCF("-1")
            gcf.bgcs = [gcf_like]

        potential_scores = []
        for spec in mf.spectra:
            spectrum_idx = self.spectra_list.index(spec)
            for bgc in gcf.bgcs:
                bgc_smiles_list = self.bgc_smiles_lookup[bgc]
                for smiles in bgc_smiles_list:
                    smiles_idx = self.smiles_list.index(smiles)
                    spectrum_smiles_score = self.iokr_scores[spectrum_idx, smiles_idx]
                    potential_scores.append(spectrum_smiles_score)

        if len(potential_scores) == 0:
            logger.info(f"No scores found for {gcf_like}")
            return 0
        else:
            return max(potential_scores)


class IOKRWrapper:
    """Wrapper around the IOKR server.
    Takes care of format conversion, fingerprint calculations, etc.
    Should also eventually take over the hardcoded stuff curently in get_iokr_server.
    """

    def __init__(self):
        self.fingerprint_type = None
        self.fingerprint_kernel = None
        self.ms_kernel = None

        self.iokr_server = None

    def _fingerprint(self, smiles):
        """Calculate molecular fingerprint for a SMILES string."""
        return fingerprint.fingerprint_from_smiles(smiles, self.fingerprint_type)

    def score_smiles(self, ms_list, candidate_smiles):
        """Score a set of spectra against a candidate set of SMILES strings."""
        spectrum_filters.datapath = get_datapath()

        logger.debug("cache miss")
        logger.debug("Calculate fingerprints for candidate set")
        # TODO: Cache this
        candidate_fps = []
        for i, c in enumerate(candidate_smiles):
            logger.debug(f"done {i}/{len(candidate_smiles)}")
            candidate_fps.append(self._fingerprint(c))
        candidate_fps = numpy.array(candidate_fps)
        logger.debug("Extract latent basis")
        latent, latent_basis, gamma = self.iokr_server.get_data_for_novel_candidate_ranking()
        logger.debug("writing cache")

        projection_matrix = numpy.zeros((len(ms_list), len(candidate_smiles)))

        # TODO: Cache this
        logger.debug("Preprocessing candidate set FPs")
        candidates = iokr_opt.preprocess_candidates(candidate_fps, latent, latent_basis, gamma)

        for ms_index, ms in enumerate(ms_list):
            logger.debug("Rank spectrum {} ({}/{})".format(ms.id, ms_index, len(ms_list)))
            ms.filter = spectrum_filters.filter_by_frozen_dag
            logger.debug("kernel vector")
            t0 = time.time()
            ms_kernel_vector = numpy.array(self.iokr_server.get_kernel_vector_for_sample(ms))
            t1 = time.time()
            logger.debug(f"done ({t1 - t0})")
            logger.debug("project")
            # projections, _ = iokr_opt.project_candidates_opt(0, candidate_fps, latent, ms_kernel_vector, latent_basis, gamma)
            projections, _ = iokr_opt.project_candidates_preprocessed(candidates, ms_kernel_vector)
            t2 = time.time()
            logger.debug(f"done ({t2 - t1})")

            logger.debug("save distances")
            projection_matrix[ms_index, :] = projections

        return projection_matrix

    def rank_smiles(self, ms, candidate_smiles):
        """Rank a spectrum against a candidate set of SMILES strings."""
        # TODO hacky
        spectrum_filters.datapath = get_datapath()
        ms.filter = spectrum_filters.filter_by_frozen_dag

        t = time.time()
        logger.debug("rank_smiles - Calculate candidate FPs")
        candidate_fps = numpy.array([self._fingerprint(x) for x in candidate_smiles])
        logger.debug(f"> {time.time() - t:.2f}s ")
        t = time.time()
        logger.debug("rank_smiles - Extract latent basis")
        latent, latent_basis, gamma = self.iokr_server.get_data_for_novel_candidate_ranking()
        logger.debug(f"> {time.time() - t:.2f}s ")
        t = time.time()
        logger.debug("rank_smiles - Get kernel vector for input sample")
        ms_kernel_vector = numpy.array(self.iokr_server.get_kernel_vector_for_sample(ms))
        logger.debug(f"> {time.time() - t:.2f}s ")
        t = time.time()
        logger.debug("rank_smiles - Rank candidate set")
        ranking, _ = iokr_opt.rank_candidates_opt(
            0, candidate_fps, latent, ms_kernel_vector, latent_basis, gamma
        )
        logger.debug(f"> {time.time() - t:.2f}s ")
        t = time.time()

        return ranking


def get_datapath():
    return os.path.join(os.path.dirname(__file__), "data")


def get_iokr_server():
    datapath = get_datapath()
    kernel_files = [
        os.path.join(datapath, f"ppk_dag_all_normalised_shifted_{x}.npy")
        for x in ["nloss", "peaks"]
    ]

    # fingerprint_type = "klekota-roth"
    fingerprint_type = None
    iokr_wrapper = IOKRWrapper()

    iokr_wrapper.fingerprint_type = fingerprint_type
    iokr_wrapper.fingerprint_kernel = None  # function

    logger.debug("Init IOKR data server")
    iokrdata = iokrdataserver.IOKRDataServer(datapath, kernel=None)

    # When the kernel is initialised from matrix, we don't have guarantee
    # that the kernel_file and iokrdata.calculate_kernel match!
    logger.debug("Init kernel values")
    # Want to be able to set this to novel kernels
    kernel_matrix = iokrdataserver.load_kernels(kernel_files)
    iokrdata.kernel = kernel_matrix

    logger.debug("Load MS files")
    iokrdata.load_ms_files(datapath)

    def ppk_wrapper(ms_i, ms_j):
        sigma_mass = 0.00001
        sigma_int = 1000000.0
        ppk_peaks = spectrum.ppk(ms_i.spectrum, ms_j.spectrum, sigma_mass, sigma_int)
        ppk_nloss = spectrum.ppk_nloss(
            ms_i.spectrum, ms_j.spectrum, ms_i.parentmass, ms_j.parentmass, sigma_mass, sigma_int
        )

        if not hasattr(ms_i, "ppk_peaks"):
            ms_i.ppk_peaks = spectrum.ppk(ms_i.spectrum, ms_i.spectrum, sigma_mass, sigma_int)
            ms_i.ppk_nloss = spectrum.ppk_nloss(
                ms_i.spectrum,
                ms_i.spectrum,
                ms_i.parentmass,
                ms_i.parentmass,
                sigma_mass,
                sigma_int,
            )

        if not hasattr(ms_j, "ppk_peaks"):
            ms_j.ppk_peaks = spectrum.ppk(ms_j.spectrum, ms_j.spectrum, sigma_mass, sigma_int)
            ms_j.ppk_nloss = spectrum.ppk_nloss(
                ms_j.spectrum,
                ms_j.spectrum,
                ms_j.parentmass,
                ms_j.parentmass,
                sigma_mass,
                sigma_int,
            )

        ppk_peaks_normalised = ppk_peaks / numpy.sqrt(ms_i.ppk_peaks * ms_j.ppk_peaks)
        ppk_nloss_normalised = ppk_nloss / numpy.sqrt(ms_i.ppk_nloss * ms_j.ppk_nloss)
        ppk = ppk_peaks_normalised + ppk_nloss_normalised / 2
        return ppk

    # The function should accept two MSSpectrum objects and return a value
    logger.debug("Configure kernel")
    iokrdata.calculate_kernel = ppk_wrapper
    # TODO: This is super slow.
    # iokrdata.build_kernel_matrix()

    logger.debug("Set fingerprint")
    iokrdata.set_fingerprint(fingerprint_type)

    all_indices = iokrdata.get_all_indices()

    iokr = iokr_opt.InputOutputKernelRegression(iokrdata)
    logger.debug("fit()")
    iokr.set_training_indices(all_indices, _lambda=0.001)
    iokr.fit()

    iokr_wrapper.iokr_server = iokr

    logger.debug("get_iokr_server complete!")
    return iokr_wrapper


def test():
    from pyteomics import mgf

    # d = mgf.read(os.path.join(datapath, 'mibig/matched_mibig_gnps_2.0.mgf'))
    d = mgf.read(os.path.join(datapath, "crusemann.mgf"))
    print(d)
    # Wrap the MGF entry in a MSSpectrum object
    test_spectrum = spectrum.MSSpectrum(d.next())

    # Candidate set
    SMILES = [
        r"C\C=C\C=C\C(=O)C1=C(O)C(C)=C(O)C(C)=C1",
        "CC1=CC2=C(C(O)=C1)C(=O)C3=C(C=C(O)C=C3O)C2=O",
        "CC1=CC2=C(C(O)=C1)C(=O)C3=C(C=C(O)C=C3O)C2=O",
        "CC1=C2C(OC(=O)C3=C2C=C(O)C=C3O)=CC(O)=C1",
        "CC1=C2C(=O)C3=C(OC2=CC(O)=C1)C=C(O)C=C3O",
        "CC1CC(C)C(=O)C(C1)C(O)CC2CC(=O)NC(=O)C2",
    ]

    iokr = get_iokr_server()
    print("done init")

    print("rank")
    rank = iokr.rank_smiles(test_spectrum, SMILES)
    print(rank)


if __name__ == "__main__":
    test()



from __future__ import annotations
import json
import logging
import re
import warnings
from collections.abc import Mapping
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
import httpx
from bs4 import BeautifulSoup
from jsonschema import validate
from nplinker.defaults import GENOME_STATUS_FILENAME
from nplinker.genomics.antismash import download_and_extract_antismash_data
from nplinker.schemas import GENOME_STATUS_SCHEMA


logger = logging.getLogger(__name__)

JGI_GENOME_LOOKUP_URL = (
    "https://img.jgi.doe.gov/cgi-bin/m/main.cgi?section=TaxonDetail&page=taxonDetail&taxon_oid={}"
)
USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:86.0) Gecko/20100101 Firefox/86.0"


class GenomeStatus:
    """Class to represent the status of a single genome.

    The status of genomes is tracked in the file
    [GENOME_STATUS_FILENAME][nplinker.defaults.GENOME_STATUS_FILENAME].
    """

    def __init__(
        self,
        original_id: str,
        resolved_refseq_id: str = "",
        resolve_attempted: bool = False,
        bgc_path: str = "",
    ):
        """Initialize a GenomeStatus object for the given genome.

        Args:
            original_id: The original ID of the genome.
            resolved_refseq_id: The resolved RefSeq ID of the
                genome. Defaults to "".
            resolve_attempted: A flag indicating whether an
                attempt to resolve the RefSeq ID has been made. Defaults to False.
            bgc_path: The path to the downloaded BGC file for
                the genome. Defaults to "".
        """
        self.original_id = original_id
        self.resolved_refseq_id = "" if resolved_refseq_id == "None" else resolved_refseq_id
        self.resolve_attempted = resolve_attempted
        self.bgc_path = bgc_path

    @staticmethod
    def read_json(file: str | PathLike) -> dict[str, "GenomeStatus"]:
        """Get a dict of GenomeStatus objects by loading given genome status file.

        Note that an empty dict is returned if the given file doesn't exist.

        Args:
            file: Path to genome status file.

        Returns:
            Dict keys are genome original id and values are GenomeStatus
                objects. An empty dict is returned if the given file doesn't exist.
        """
        genome_status_dict = {}
        if Path(file).exists():
            with open(file, "r") as f:
                data = json.load(f)

            # validate json data before using it
            validate(data, schema=GENOME_STATUS_SCHEMA)

            genome_status_dict = {
                gs["original_id"]: GenomeStatus(**gs) for gs in data["genome_status"]
            }
        return genome_status_dict

    @staticmethod
    def to_json(
        genome_status_dict: Mapping[str, "GenomeStatus"], file: str | PathLike | None = None
    ) -> str | None:
        """Convert the genome status dictionary to a JSON string.

        If a file path is provided, the JSON string is written to the file. If
        the file already exists, it is overwritten.

        Args:
            genome_status_dict: A dictionary of genome
                status objects. The keys are the original genome IDs and the values
                are GenomeStatus objects.
            file: The path to the output JSON file.
                If None, the JSON string is returned but not written to a file.

        Returns:
            The JSON string if `file` is None, otherwise None.
        """
        gs_list = [gs._to_dict() for gs in genome_status_dict.values()]
        json_data = {"genome_status": gs_list, "version": "1.0"}

        # validate json object before dumping
        validate(json_data, schema=GENOME_STATUS_SCHEMA)

        if file is not None:
            with open(file, "w") as f:
                json.dump(json_data, f)
            return None
        return json.dumps(json_data)

    def _to_dict(self) -> dict:
        """Convert the GenomeStatus object to a dict."""
        return {
            "original_id": self.original_id,
            "resolved_refseq_id": self.resolved_refseq_id,
            "resolve_attempted": self.resolve_attempted,
            "bgc_path": self.bgc_path,
        }


def podp_download_and_extract_antismash_data(
    genome_records: Sequence[Mapping[str, Mapping[str, str]]],
    project_download_root: str | PathLike,
    project_extract_root: str | PathLike,
):
    """Download and extract antiSMASH BGC archive for the given genome records.

    Args:
        genome_records: list of dicts representing genome records.

            The dict of each genome record contains a key of genome ID with a value
            of another dict containing information about genome type, label and
            accession ids (RefSeq, GenBank, and/or JGI).
        project_download_root: Path to the directory to place
            downloaded archive in.
        project_extract_root: Path to the directory downloaded archive will be extracted to.

            Note that an `antismash` directory will be created in the specified
            `extract_root` if it doesn't exist. The files will be extracted to
            `<extract_root>/antismash/<antismash_id>` directory.

    Warnings:
        UserWarning: when no antiSMASH data is found for some genomes.
    """
    if not Path(project_download_root).exists():
        # otherwise in case of failed first download, the folder doesn't exist and
        # genome_status_file can't be written
        Path(project_download_root).mkdir(parents=True, exist_ok=True)

    gs_file = Path(project_download_root, GENOME_STATUS_FILENAME)
    gs_dict = GenomeStatus.read_json(gs_file)

    for i, genome_record in enumerate(genome_records):
        # get the best available ID from the dict
        genome_id_data = genome_record["genome_ID"]
        raw_genome_id = get_best_available_genome_id(genome_id_data)
        if raw_genome_id is None or len(raw_genome_id) == 0:
            logger.warning(f'Invalid input genome record "{genome_record}"')
            continue

        # check if genome ID exist in the genome status file
        if raw_genome_id not in gs_dict:
            gs_dict[raw_genome_id] = GenomeStatus(raw_genome_id)

        gs_obj = gs_dict[raw_genome_id]

        logger.info(
            f"Checking for antismash data {i + 1}/{len(genome_records)}, "
            f"current genome ID={raw_genome_id}"
        )
        # first, check if BGC data is downloaded
        if gs_obj.bgc_path and Path(gs_obj.bgc_path).exists():
            logger.info(f"Genome ID {raw_genome_id} already downloaded to {gs_obj.bgc_path}")
            continue
        # second, check if lookup attempted previously
        if gs_obj.resolve_attempted:
            logger.info(f"Genome ID {raw_genome_id} skipped due to previous failed attempt")
            continue

        # if not downloaded or lookup attempted, then try to resolve the ID
        # and download
        logger.info(f"Start lookup process for genome ID {raw_genome_id}")
        gs_obj.resolved_refseq_id = _resolve_refseq_id(genome_id_data)
        gs_obj.resolve_attempted = True

        if gs_obj.resolved_refseq_id == "":
            # give up on this one
            logger.warning(f"Failed lookup for genome ID {raw_genome_id}")
            continue

        # if resolved id is valid, try to download and extract antismash data
        try:
            download_and_extract_antismash_data(
                gs_obj.resolved_refseq_id, project_download_root, project_extract_root
            )

            gs_obj.bgc_path = str(
                Path(project_download_root, gs_obj.resolved_refseq_id + ".zip").absolute()
            )

            output_path = Path(project_extract_root, "antismash", gs_obj.resolved_refseq_id)
            if output_path.exists():
                Path.touch(output_path / "completed", exist_ok=True)

        except Exception:
            gs_obj.bgc_path = ""

    # raise and log warning for failed downloads
    failed_ids = [gs.original_id for gs in gs_dict.values() if not gs.bgc_path]
    if failed_ids:
        warning_message = (
            f"Failed to download antiSMASH data for the following genome IDs: {failed_ids}"
        )
        logger.warning(warning_message)
        warnings.warn(warning_message, UserWarning)

    # save updated genome status to json file
    GenomeStatus.to_json(gs_dict, gs_file)

    if len(failed_ids) == len(genome_records):
        raise ValueError("No antiSMASH data found for any genome")


def get_best_available_genome_id(genome_id_data: Mapping[str, str]) -> str | None:
    """Get the best available ID from genome_id_data dict.

    Args:
        genome_id_data: dictionary containing information for each genome record present.

    Returns:
        ID for the genome, if present, otherwise None.
    """
    if "RefSeq_accession" in genome_id_data:
        best_id = genome_id_data["RefSeq_accession"]
    elif "GenBank_accession" in genome_id_data:
        best_id = genome_id_data["GenBank_accession"]
    elif "JGI_Genome_ID" in genome_id_data:
        best_id = genome_id_data["JGI_Genome_ID"]
    else:
        best_id = None

    if best_id is None or len(best_id) == 0:
        logger.warning(f"Failed to get valid genome ID in genome data: {genome_id_data}")
        return None
    return best_id


def _resolve_genbank_accession(genbank_id: str) -> str:
    """Try to get RefSeq assembly id through given GenBank assembly id.

    Note that GenBank assembly accession starts with "GCA_" and RefSeq assembly
    accession starts with "GCF_". For more info, see
    https://www.ncbi.nlm.nih.gov/datasets/docs/v2/troubleshooting/faq

    Args:
        genbank_id: ID for GenBank assembly accession.

    Raises:
        httpx.ReadTimeout: If the request times out.

    Returns:
        RefSeq assembly ID if the search is successful, otherwise an empty string.
    """
    logger.info(
        f"Attempting to resolve Genbank assembly accession {genbank_id} to RefSeq accession"
    )
    # NCBI Datasets API https://www.ncbi.nlm.nih.gov/datasets/docs/v2/api/
    # Note that there is a API rate limit of 5 requests per second without using an API key
    # For more info, see https://www.ncbi.nlm.nih.gov/datasets/docs/v2/troubleshooting/faq/

    # API for getting revision history of a genome assembly
    # For schema, see https://www.ncbi.nlm.nih.gov/datasets/docs/v2/api/rest-api/#get-/genome/accession/-accession-/revision_history
    url = f"https://api.ncbi.nlm.nih.gov/datasets/v2/genome/accession/{genbank_id}/revision_history"

    refseq_id = ""
    try:
        resp = httpx.get(
            url, headers={"User-Agent": USER_AGENT}, timeout=10.0, follow_redirects=True
        )
        if resp.status_code == httpx.codes.OK:
            data = resp.json()
            latest_entry = max(
                (entry for entry in data["assembly_revisions"] if "refseq_accession" in entry),
                key=lambda x: x["release_date"],
            )
            refseq_id = latest_entry["refseq_accession"]
    except httpx.ReadTimeout:
        logger.warning("Timed out waiting for result of GenBank assembly lookup")

    return refseq_id


def _resolve_jgi_accession(jgi_id: str) -> str:
    """Try to get RefSeq id through given JGI id.

    Args:
        jgi_id: JGI_Genome_ID for GenBank accession.

    Returns:
        RefSeq ID if search is successful, otherwise an empty string.
    """
    url = JGI_GENOME_LOOKUP_URL.format(jgi_id)
    logger.info(f"Attempting to resolve JGI_Genome_ID {jgi_id} to GenBank accession via {url}")
    # no User-Agent header produces a 403 Forbidden error on this site...
    try:
        resp = httpx.get(
            url, headers={"User-Agent": USER_AGENT}, timeout=10.0, follow_redirects=True
        )
    except httpx.ReadTimeout:
        logger.warning("Timed out waiting for result of JGI_Genome_ID lookup")
        return ""

    soup = BeautifulSoup(resp.content, "html.parser")
    # Find the table entry giving the "NCBI Assembly Accession" ID
    link = soup.find("a", href=re.compile("https://www.ncbi.nlm.nih.gov/datasets/genome/.*"))
    if link is None:
        return ""

    assembly_id = link.text
    # check if the assembly ID is already a RefSeq ID
    if assembly_id.startswith("GCF_"):
        return assembly_id
    else:
        return _resolve_genbank_accession(assembly_id)


def _resolve_refseq_id(genome_id_data: Mapping[str, str]) -> str:
    """Get the RefSeq ID to which the genome accession is linked.

    Check https://pairedomicsdata.bioinformatics.nl/schema.json.

    Args:
        genome_id_data: dictionary containing information
        for each genome record present.

    Returns:
        RefSeq ID if present, otherwise an empty string.
    """
    if "RefSeq_accession" in genome_id_data:
        # best case, can use this directly
        return genome_id_data["RefSeq_accession"]
    if "GenBank_accession" in genome_id_data:
        # resolve via NCBI
        return _resolve_genbank_accession(genome_id_data["GenBank_accession"])
    if "JGI_Genome_ID" in genome_id_data:
        # resolve via JGI => NCBI
        return _resolve_jgi_accession(genome_id_data["JGI_Genome_ID"])

    logger.warning(f"Unable to resolve genome_ID: {genome_id_data}")
    return ""



from __future__ import annotations
import fnmatch
import logging
import os
from collections.abc import Mapping
from os import PathLike
from pathlib import Path
from Bio import SeqIO
from Bio import SeqRecord
from nplinker.genomics import BGC
from nplinker.strain import Strain
from nplinker.utils import list_dirs
from nplinker.utils import list_files
from ..abc import BGCLoaderBase


logger = logging.getLogger(__name__)


class AntismashBGCLoader(BGCLoaderBase):
    """Data loader for AntiSMASH BGC genbank (.gbk) files."""

    def __init__(self, data_dir: str | PathLike) -> None:
        """Initialize the AntiSMASH BGC loader.

        Args:
            data_dir: Path to AntiSMASH directory that contains a collection of AntiSMASH outputs.

        Notes:
            The input `data_dir` must follow the structure defined in the
            [Working Directory Structure][working-directory-structure] for AntiSMASH data, e.g.:
            ```shell
            antismash
                ├── genome_id_1                  # one AntiSMASH output, e.g. GCF_000514775.1
                │  ├── NZ_AZWO01000004.region001.gbk
                │  └── ...
                ├── genome_id_2
                │  ├── ...
                └── ...
            ```
        """
        self.data_dir = str(data_dir)
        self._file_dict = self._parse_data_dir(self.data_dir)
        self._bgcs = self._parse_bgcs(self._file_dict)

    def get_bgc_genome_mapping(self) -> dict[str, str]:
        """Get the mapping from BGC to genome.

        !!! info
            The directory name of the gbk files is treated as genome id.

        Returns:
            The key is BGC name (gbk file name) and value is genome id (the directory name of the
            gbk file).
        """
        return {
            bid: os.path.basename(os.path.dirname(bpath)) for bid, bpath in self._file_dict.items()
        }

    def get_files(self) -> dict[str, str]:
        """Get BGC gbk files.

        Returns:
            The key is BGC name (gbk file name) and value is path to the gbk file.
        """
        return self._file_dict

    @staticmethod
    def _parse_data_dir(data_dir: str) -> dict[str, str]:
        """Parse AntiSMASH directory to get path of all BGC gbk files.

        Args:
            data_dir: Path to AntiSMASH directory that contains
                a collection of AntiSMASH outputs

        Returns:
            The key is BGC name (gbk file name) and value is path to the gbk file.
        """
        bgc_files = {}
        subdirs = list_dirs(data_dir)
        for subdir in subdirs:
            # get all .gbk files
            files = list_files(subdir, suffix=".gbk", keep_parent=False)
            # filter BGC's .gbk files
            files = fnmatch.filter(files, "*.region???.gbk")
            for f in files:
                fname = os.path.splitext(f)[0]
                fpath = os.path.join(subdir, f)
                bgc_files[fname] = fpath

        return bgc_files

    def get_bgcs(self) -> list[BGC]:
        """Get all BGC objects.

        Returns:
            A list of BGC objects
        """
        return self._bgcs

    @staticmethod
    def _parse_bgcs(bgc_files: Mapping[str, str]) -> list[BGC]:
        """Load given BGC files as BGC objects.

        Args:
            bgc_files: key is BGC name and value is path to the
                BGC gbk file, see method :meth:`.bgc_files`.

        Returns:
            A list of BGC objects
        """
        return [parse_bgc_genbank(file) for file in bgc_files.values()]


def parse_bgc_genbank(file: str | PathLike) -> BGC:
    """Parse a single BGC gbk file to BGC object.

    Args:
        file: Path to BGC gbk file

    Returns:
        BGC object

    Examples:
        >>> bgc = AntismashBGCLoader.parse_bgc(
        ...    "/data/antismash/GCF_000016425.1/NC_009380.1.region001.gbk")
    """
    file = Path(file)
    fname = file.stem

    record = SeqIO.read(file, format="genbank")
    description = record.description  # "DEFINITION" in gbk file
    antismash_id = record.id  # "VERSION" in gbk file
    features = _parse_antismash_genbank(record)
    product_prediction = features.get("product")
    if product_prediction is None:
        raise ValueError(f"Not found product prediction in antiSMASH Genbank file {file}")

    # init BGC
    bgc = BGC(fname, *product_prediction)
    bgc.description = description
    bgc.antismash_id = antismash_id
    bgc.antismash_file = str(file)
    bgc.antismash_region = features.get("region_number")
    bgc.smiles = features.get("smiles")
    bgc.strain = Strain(fname)
    return bgc


def _parse_antismash_genbank(record: SeqRecord.SeqRecord) -> dict:
    features = {}
    for feature in record.features:
        if feature.type == "region":
            # biopython assumes region numer is a list, but it's actually an int
            features["region_number"] = feature.qualifiers.get("region_number")[0]
            features["product"] = feature.qualifiers.get("product")
        if feature.type == "cand_cluster":
            smiles = feature.qualifiers.get("SMILES")
            # space is not allowed in SMILES spec
            # biopython generates space when reading multi-line SMILES from .gbk
            if smiles is not None:
                smiles = tuple(i.replace(" ", "") for i in smiles)
            features["smiles"] = smiles
    return features



from __future__ import annotations
import logging
import os
import subprocess
import sys
from os import PathLike
from typing import Literal


logger = logging.getLogger(__name__)

PFAM_PATH = os.path.join(sys.prefix, "nplinker_lib")


def run_bigscape(
    antismash_path: str | PathLike,
    output_path: str | PathLike,
    extra_params: str,
    version: Literal[1, 2] = 1,
) -> bool:
    """Runs BiG-SCAPE to cluster BGCs.

    The behavior of this function is slightly different depending on the version of
    BiG-SCAPE that is set to run using the configuration file.
    Mostly this means a different set of parameters is used between the two versions.

    The AntiSMASH output directory should be a directory that contains GBK files.
    The directory can contain subdirectories, in which case BiG-SCAPE will search
    recursively for GBK files. E.g.:

    ```
    example_folder
        ├── organism_1
        │  ├── organism_1.region001.gbk
        │  ├── organism_1.region002.gbk
        │  ├── organism_1.region003.gbk
        │  ├── organism_1.final.gbk          <- skipped!
        │  └── ...
        ├── organism_2
        │  ├── ...
        └── ...
    ```

    By default, only GBK Files with "cluster" or "region" in the filename are
    accepted. GBK Files with "final" in the filename are excluded.

    Args:
        antismash_path: Path to the antismash output directory.
        output_path: Path to the output directory where BiG-SCAPE will write its results.
        extra_params: Additional parameters to pass to BiG-SCAPE.
        version: The version of BiG-SCAPE to run. Must be 1 or 2.

    Returns:
        True if BiG-SCAPE ran successfully, False otherwise.

    Raises:
        ValueError: If an unexpected BiG-SCAPE version number is specified.
        FileNotFoundError: If the antismash_path does not exist or if the BiG-SCAPE python
            script could not be found.
        RuntimeError: If BiG-SCAPE fails to run.

    Examples:
        >>>  from nplinker.genomics.bigscape import run_bigscape
        >>> run_bigscape(antismash_path="./antismash", output_path="./output",
        ... extra_params="--help", version=1)
    """
    # switch to correct version of BiG-SCAPE
    if version == 1:
        bigscape_py_path = "bigscape.py"
    elif version == 2:
        bigscape_py_path = "bigscape-v2.py"
    else:
        raise ValueError("Invalid BiG-SCAPE version number. Expected: 1 or 2.")

    try:
        subprocess.run([bigscape_py_path, "-h"], capture_output=True, check=True)
    except Exception as e:
        raise FileNotFoundError(
            f"Failed to find/run BiG-SCAPE executable program (path={bigscape_py_path}, err={e})"
        ) from e

    if not os.path.exists(antismash_path):
        raise FileNotFoundError(f'antismash_path "{antismash_path}" does not exist!')

    logger.info(f"Running BiG-SCAPE version {version}")
    logger.info(
        f'run_bigscape: input="{antismash_path}", output="{output_path}", extra_params={extra_params}"'
    )

    # assemble arguments. first argument is the python file
    args = [bigscape_py_path]

    # version 2 points to specific Pfam file, version 1 points to directory
    # version 2 also requires the cluster subcommand
    if version == 1:
        args.extend(["--pfam_dir", PFAM_PATH])
    elif version == 2:
        args.extend(["cluster", "--pfam_path", os.path.join(PFAM_PATH, "Pfam-A.hmm")])

    # add input and output paths. these are unchanged
    args.extend(["-i", str(antismash_path), "-o", str(output_path)])

    # append the user supplied params, if any
    if len(extra_params) > 0:
        args.extend(extra_params.split(" "))

    logger.info(f"BiG-SCAPE command: {args}")
    result = subprocess.run(args, stdout=sys.stdout, stderr=sys.stderr)

    # return true on any non-error return code
    if result.returncode == 0:
        logger.info(f"BiG-SCAPE completed with return code {result.returncode}")
        return True

    # otherwise log details and raise a runtime error
    logger.error(f"BiG-SCAPE failed with return code {result.returncode}")
    logger.error(f"output: {str(result.stdout)}")
    logger.error(f"stderr: {str(result.stderr)}")

    raise RuntimeError(f"Failed to run BiG-SCAPE with error code {result.returncode}")



from __future__ import annotations
import logging
import pickle
from collections.abc import Sequence
from os import PathLike
from pprint import pformat
from typing import Any
from typing import overload
from dynaconf import Dynaconf
from .arranger import DatasetArranger
from .config import load_config
from .defaults import OUTPUT_DIRNAME
from .genomics import BGC
from .genomics import GCF
from .loader import DatasetLoader
from .logger import setup_logging
from .metabolomics import MolecularFamily
from .metabolomics import Spectrum
from .scoring.link_graph import LinkGraph
from .scoring.metcalf_scoring import MetcalfScoring
from .strain import StrainCollection


logger = logging.getLogger(__name__)


class NPLinker:
    """The central class of NPLinker application.

    Attributes:
        config: The configuration object for the current NPLinker application.
        root_dir: The path to the root directory of the current NPLinker application.
        output_dir: The path to the output directory of the current NPLinker application.
        bgcs: A list of all BGC objects.
        gcfs: A list of all GCF objects.
        spectra: A list of all Spectrum objects.
        mfs: A list of all MolecularFamily objects.
        mibig_bgcs: A list of all MiBIG BGC objects.
        strains: A StrainCollection object containing all Strain objects.
        product_types: A list of all BiGSCAPE product types.
        scoring_methods: A list of all valid scoring methods.
    """

    # Valid scoring methods
    _valid_scoring_methods = {
        MetcalfScoring.name: MetcalfScoring,
        # RosettaScoring.name: RosettaScoring, # To be refactored
        # NPClassScoring.name: NPClassScoring, # To be refactored
    }

    def __init__(self, config_file: str | PathLike):
        """Initialise an NPLinker instance.

        Args:
            config_file: Path to the configuration file to use.


        Examples:
            Starting the NPLinker application:
            >>> from nplinker import NPLinker
            >>> npl = NPLinker("path/to/config.toml")

            Loading data from files to python objects:
            >>> npl.load_data()

            Checking the number of GCF objects:
            >>> len(npl.gcfs)

            Getting the links for all GCF objects using the Metcalf scoring method, and the result
            is stored in a [LinkGraph][nplinker.scoring.LinkGraph] object:
            >>> lg = npl.get_links(npl.gcfs, "metcalf")

            Getting the link data between two objects:
            >>> link_data = lg.get_link_data(npl.gcfs[0], npl.spectra[0])
            {"metcalf": Score("metcalf", 1.0, {"cutoff": 0, "standardised": False})}

            Saving the data to a pickle file:
            >>> npl.save_data("path/to/output.pkl", lg)
        """
        # Load the configuration file
        self.config: Dynaconf = load_config(config_file)

        # Setup logging for the application
        setup_logging(
            level=self.config.log.level,
            file=self.config.log.get("file", ""),
            use_console=self.config.log.use_console,
        )
        logger.info(
            "Configuration:\n %s", pformat(self.config.as_dict(), width=20, sort_dicts=False)
        )

        # Setup the output directory
        self._output_dir = self.config.root_dir / OUTPUT_DIRNAME
        self._output_dir.mkdir(exist_ok=True)

        # Initialise data containers that will be populated by the `load_data` method
        self._bgc_dict: dict[str, BGC] = {}
        self._gcf_dict: dict[str, GCF] = {}
        self._spec_dict: dict[str, Spectrum] = {}
        self._mf_dict: dict[str, MolecularFamily] = {}
        self._mibig_bgcs: list[BGC] = []
        self._strains: StrainCollection = StrainCollection()
        self._product_types: list = []
        self._chem_classes = None  # TODO: to be refactored
        self._class_matches = None  # TODO: to be refactored

        # Flags to keep track of whether the scoring methods have been set up
        self._scoring_methods_setup_done = {name: False for name in self._valid_scoring_methods}

    @property
    def root_dir(self) -> str:
        """Get the path to the root directory of the current NPLinker instance."""
        return str(self.config.root_dir)

    @property
    def output_dir(self) -> str:
        """Get the path to the output directory of the current NPLinker instance."""
        return str(self._output_dir)

    @property
    def bgcs(self) -> list[BGC]:
        """Get all BGC objects."""
        return list(self._bgc_dict.values())

    @property
    def gcfs(self) -> list[GCF]:
        """Get all GCF objects."""
        return list(self._gcf_dict.values())

    @property
    def spectra(self) -> list[Spectrum]:
        """Get all Spectrum objects."""
        return list(self._spec_dict.values())

    @property
    def mfs(self) -> list[MolecularFamily]:
        """Get all MolecularFamily objects."""
        return list(self._mf_dict.values())

    @property
    def mibig_bgcs(self) -> list[BGC]:
        """Get all MiBIG BGC objects."""
        return self._mibig_bgcs

    @property
    def strains(self) -> StrainCollection:
        """Get all Strain objects."""
        return self._strains

    @property
    def product_types(self) -> list[str]:
        """Get all BiGSCAPE product types."""
        return self._product_types

    @property
    def chem_classes(self):
        """Returns loaded ChemClassPredictions with the class predictions."""
        return self._chem_classes

    @property
    def class_matches(self):
        """ClassMatches with the matched classes and scoring tables from MIBiG."""
        return self._class_matches

    @property
    def scoring_methods(self) -> list[str]:
        """Get names of all valid scoring methods."""
        return list(self._valid_scoring_methods.keys())

    def load_data(self):
        """Load all data from files into memory.

        This method is a convenience function that calls the
        [`DatasetArranger`][nplinker.arranger.DatasetArranger] class to arrange data files
        (download, generate and/or validate data) in the [correct directory structure][working-directory-structure],
        and then calls the [`DatasetLoader`][nplinker.loader.DatasetLoader] class to load all data
        from the files into memory.

        The loaded data is stored in various data containers for easy access, e.g.
        [`self.bgcs`][nplinker.NPLinker.bgcs] for all BGC objects,
        [`self.strains`][nplinker.NPLinker.strains] for all Strain objects, etc.
        """
        arranger = DatasetArranger(self.config)
        arranger.arrange()
        loader = DatasetLoader(self.config)
        loader.load()

        self._bgc_dict = {bgc.id: bgc for bgc in loader.bgcs}
        self._gcf_dict = {gcf.id: gcf for gcf in loader.gcfs}
        self._spec_dict = {spec.id: spec for spec in loader.spectra}
        self._mf_dict = {mf.id: mf for mf in loader.mfs}

        self._mibig_bgcs = loader.mibig_bgcs
        self._strains = loader.strains
        self._product_types = loader.product_types
        self._chem_classes = loader.chem_classes
        self._class_matches = loader.class_matches

    @overload
    def get_links(
        self, objects: Sequence[BGC], scoring_method: str, **scoring_params: Any
    ) -> LinkGraph: ...
    @overload
    def get_links(
        self, objects: Sequence[GCF], scoring_method: str, **scoring_params: Any
    ) -> LinkGraph: ...
    @overload
    def get_links(
        self, objects: Sequence[Spectrum], scoring_method: str, **scoring_params: Any
    ) -> LinkGraph: ...
    @overload
    def get_links(
        self, objects: Sequence[MolecularFamily], scoring_method: str, **scoring_params: Any
    ) -> LinkGraph: ...

    def get_links(
        self,
        objects: Sequence[BGC] | Sequence[GCF] | Sequence[Spectrum] | Sequence[MolecularFamily],
        scoring_method: str,
        **scoring_params: Any,
    ) -> LinkGraph:
        """Get links for the given objects using the specified scoring method and parameters.

        Args:
            objects: A sequence of objects to get links for. The objects must be of the same
                type, i.e. `BGC`, `GCF`, `Spectrum` or `MolecularFamily` type.
                !!! Warning
                    For scoring method `metcalf`, the `BGC` objects are not supported.
            scoring_method: The scoring method to use. Must be one of the valid scoring methods
                [`self.scoring_methods`][nplinker.NPLinker.scoring_methods], such as `metcalf`.
            scoring_params: Parameters to pass to the scoring method. If not given, the default
                parameters of the specified scoring method will be used.

                Check the `get_links` method of the scoring method class for the available
                parameters and their default values.

                | Scoring Method | Scoring Parameters |
                | -------------- | ------------------ |
                | `metcalf` | [`cutoff`, `standardised`][nplinker.scoring.MetcalfScoring.get_links] |

        Returns:
            A LinkGraph object containing the links for the given objects.

        Raises:
            ValueError: If input objects are empty or if the scoring method is invalid.
            TypeError: If the input objects are not of the same type or if the object type is invalid.

        Examples:
            Using default scoring parameters:
            >>> lg = npl.get_links(npl.gcfs, "metcalf")

            Scoring parameters provided:
            >>> lg = npl.get_links(npl.gcfs, "metcalf", cutoff=0.5, standardised=True)
        """
        # Validate objects
        if len(objects) == 0:
            raise ValueError("No objects provided to get links for")
        # check if all objects are of the same type
        types = {type(i) for i in objects}
        if len(types) > 1:
            raise TypeError("Input objects must be of the same type.")
        # check if the object type is valid
        obj_type = next(iter(types))
        if obj_type not in (BGC, GCF, Spectrum, MolecularFamily):
            raise TypeError(
                f"Invalid type {obj_type}. Input objects must be BGC, GCF, Spectrum or MolecularFamily objects."
            )

        # Validate scoring method
        if scoring_method not in self._valid_scoring_methods:
            raise ValueError(f"Invalid scoring method {scoring_method}.")

        # Check if the scoring method has been set up
        if not self._scoring_methods_setup_done[scoring_method]:
            self._valid_scoring_methods[scoring_method].setup(self)
            self._scoring_methods_setup_done[scoring_method] = True

        # Initialise the scoring method
        scoring = self._valid_scoring_methods[scoring_method]()

        return scoring.get_links(*objects, **scoring_params)

    def lookup_bgc(self, id: str) -> BGC | None:
        """Get the BGC object with the given ID.

        Args:
            id: the ID of the BGC to look up.

        Returns:
            The BGC object with the given ID, or None if no such object exists.

        Examples:
            >>> bgc = npl.lookup_bgc("BGC000001")
            >>> bgc
            BGC(id="BGC000001", ...)
        """
        return self._bgc_dict.get(id, None)

    def lookup_gcf(self, id: str) -> GCF | None:
        """Get the GCF object with the given ID.

        Args:
            id: the ID of the GCF to look up.

        Returns:
            The GCF object with the given ID, or None if no such object exists.
        """
        return self._gcf_dict.get(id, None)

    def lookup_spectrum(self, id: str) -> Spectrum | None:
        """Get the Spectrum object with the given ID.

        Args:
            id: the ID of the Spectrum to look up.

        Returns:
            The Spectrum object with the given ID, or None if no such object exists.
        """
        return self._spec_dict.get(id, None)

    def lookup_mf(self, id: str) -> MolecularFamily | None:
        """Get the MolecularFamily object with the given ID.

        Args:
            id: the ID of the MolecularFamily to look up.

        Returns:
            The MolecularFamily object with the given ID, or None if no such object exists.
        """
        return self._mf_dict.get(id, None)

    def save_data(
        self,
        file: str | PathLike,
        links: LinkGraph | None = None,
    ) -> None:
        """Pickle data to a file.

        The  pickled data is a tuple of BGCs, GCFs, Spectra, MolecularFamilies, StrainCollection and
        links, i.e. `(bgcs, gcfs, spectra, mfs, strains, links)`.

        Args:
            file: The path to the pickle file to save the data to.
            links: The LinkGraph object to save.

        Examples:
            Saving the data to a pickle file, links data is `None`:
            >>> npl.save_data("path/to/output.pkl")

            Also saving the links data:
            >>> lg = npl.get_links(npl.gcfs, "metcalf")
            >>> npl.save_data("path/to/output.pkl", lg)
        """
        data = (self.bgcs, self.gcfs, self.spectra, self.mfs, self.strains, links)
        with open(file, "wb") as f:
            pickle.dump(data, f)



from __future__ import annotations
import csv
import logging
import sqlite3
from os import PathLike
from ..abc import GCFLoaderBase
from ..gcf import GCF


logger = logging.getLogger(__name__)


class BigscapeGCFLoader(GCFLoaderBase):
    """Data loader for BiG-SCAPE GCF cluster file.

    Attributes:
        cluster_file: path to the BiG-SCAPE cluster file.
    """

    def __init__(self, cluster_file: str | PathLike, /) -> None:
        """Initialize the BiG-SCAPE GCF loader.

        Args:
            cluster_file: Path to the BiG-SCAPE cluster file,
                the filename has a pattern of `<class>_clustering_c0.xx.tsv`.
        """
        self.cluster_file: str = str(cluster_file)
        self._gcf_list = self._parse_gcf(self.cluster_file)

    def get_gcfs(self, keep_mibig_only: bool = False, keep_singleton: bool = False) -> list[GCF]:
        """Get all GCF objects.

        Args:
            keep_mibig_only: True to keep GCFs that contain only MIBiG
                BGCs.
            keep_singleton: True to keep singleton GCFs. A singleton GCF
                is a GCF that contains only one BGC.

        Returns:
            A list of GCF objects.
        """
        gcf_list = self._gcf_list
        if not keep_mibig_only:
            gcf_list = [gcf for gcf in gcf_list if not gcf.has_mibig_only()]
        if not keep_singleton:
            gcf_list = [gcf for gcf in gcf_list if not gcf.is_singleton()]
        return gcf_list

    @staticmethod
    def _parse_gcf(cluster_file: str) -> list[GCF]:
        """Parse BiG-SCAPE cluster file to return GCF objects."""
        gcf_dict: dict[str, GCF] = {}
        with open(cluster_file, "rt", encoding="utf-8") as f:
            reader = csv.reader(f, delimiter="\t")
            next(reader)  # skip headers
            for line in reader:
                bgc_id, family_id = line[:]
                if family_id not in gcf_dict:
                    gcf_dict[family_id] = GCF(family_id)
                gcf_dict[family_id].bgc_ids.add(bgc_id)
        return list(gcf_dict.values())


class BigscapeV2GCFLoader(GCFLoaderBase):
    """Data loader for BiG-SCAPE v2 database file.

    Attributes:
        db_file: Path to the BiG-SCAPE database file.
    """

    def __init__(self, db_file: str | PathLike, /) -> None:
        """Initialize the BiG-SCAPE v2 GCF loader.

        Args:
            db_file: Path to the BiG-SCAPE v2 database file
        """
        self.db_file = str(db_file)
        self._gcf_list = self._parse_gcf(self.db_file)

    def get_gcfs(self, keep_mibig_only: bool = False, keep_singleton: bool = False) -> list[GCF]:
        """Get all GCF objects.

        Args:
            keep_mibig_only: True to keep GCFs that contain only MIBiG BGCs.
            keep_singleton: True to keep singleton GCFs.
                A singleton GCF is a GCF that contains only one BGC.

        Returns:
            a list of GCF objects.
        """
        gcf_list = self._gcf_list
        if not keep_mibig_only:
            gcf_list = [gcf for gcf in gcf_list if not gcf.has_mibig_only()]
        if not keep_singleton:
            gcf_list = [gcf for gcf in gcf_list if not gcf.is_singleton()]
        return gcf_list

    @staticmethod
    def _parse_gcf(db_file: str) -> list[GCF]:
        """Get GCF objects from database.

        Args:
            db_file: Path to the sqlite3 database file.

        Returns:
            A list of GCF objects
        """
        gcf_dict: dict[str, GCF] = {}

        with sqlite3.connect(db_file) as connection:
            cursor = connection.cursor()

            query = """
            SELECT gbk.path, bgc_record_family.family_id FROM bgc_record_family
            JOIN bgc_record ON bgc_record.id = bgc_record_family.record_id
            JOIN gbk ON gbk.id = bgc_record.gbk_id
            """

            results = cursor.execute(query).fetchall()

            for result in results:
                gbk_path, family_id = result

                # take the filename of the gbk path as the bgc_id
                # filename
                bgc_id: str = gbk_path.split("/")[-1]
                # remove extension
                bgc_id = bgc_id.rsplit(".", 1)[0]

                if family_id not in gcf_dict:
                    gcf_dict[family_id] = GCF(family_id)
                gcf_dict[family_id].bgc_ids.add(bgc_id)

        return list(gcf_dict.values())



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob
import logging
import os
from collections import Counter
from canopus import Canopus
from canopus.classifications_to_gnps import analyse_canopus


logger = logging.getLogger(__name__)


# load Chem_class_predictions (canopus, molnetenhancer are loaded)
# for canopus, check if results can be converted with canopus_treemap
# otherwise use the pre-existing output of canopus
class ChemClassPredictions:
    """Class for storing results for chemical class predictions of spectra.

    Currently, CANOPUS and MolNetEnhancer results are loaded
    """

    def __init__(self, canopus_dir, mne_dir, gnps_dir):
        """Load classes with CanopusResults, MolNetEnhancerResults.

        Args:
            canopus_dir: str, canopus_dir found in root_dir of nplinker project
            mne_dir: str, mne_dir found in root_dir of nplinker project
            gnps_dir: str, root dir where all gnps info is found
        """
        self._canopus = CanopusResults(canopus_dir, gnps_dir)
        self._molnetenhancer = MolNetEnhancerResults(mne_dir)

        class_predict_options = []
        if self._canopus.spectra_classes:
            class_predict_options.append("canopus")
        if self._molnetenhancer.spectra2mf:
            class_predict_options.append("molnetenhancer")
        if class_predict_options:
            class_predict_options = ["mix", "main"] + class_predict_options
        self._class_predict_options = class_predict_options

    @property
    def canopus(self):
        return self._canopus

    @property
    def molnetenhancer(self):
        return self._molnetenhancer

    @property
    def class_predict_options(self):
        """The available class predictions."""
        return self._class_predict_options


class CanopusResults:
    """Class for storing canopus results.

    The results from the canopus dir are read and combined with the MN from GNPS
    using canopus_treemap: github.com/louwenjjr/canopus_treemap/tree/master/canopus
    This creates the two files that are read for the spectra and mfs:
        -cluster_index_classifications.txt
        -component_index_classifications.txt

    If canopus_treemap somehow fails (like for too old gnps results), the canopus
    output is read as is from canopus_summary.tsv
    """

    def __init__(self, canopus_dir, gnps_dir):
        """Read the class info from root_dir/canopus.

        Args:
            canopus_dir: str, canopus_dir found in root_dir of nplinker project
            gnps_dir: str, root dir where all gnps info is found

        If possible, convert canopus output with canopus_treemap. Otherwise
        canopus output is used directly. If converting fails, probably the MN
        version is too old, and canopus output is read directly
        """
        self._canopus_dir = canopus_dir
        self._gnps_dir = gnps_dir
        self._mf_classes, self._mf_classes_names, self._mf_classes_names_inds = (
            None,
            None,
            None,
        )
        self._spectra_classes, self._spectra_classes_names, self._spectra_classes_names_inds = (
            None,
            None,
            None,
        )
        if os.path.isdir(self._canopus_dir):
            self._read_all_classes()
        else:
            logger.info(
                f"No CANOPUS results present at {self._canopus_dir}. "
                f"(set run_canopus=true in the .toml to run CANOPUS)"
            )

    def _read_all_classes(self):
        """Wrapper to read all canopus output and store them in the object."""
        ci_file = False  # find cluster_index_classifications in canopus_dir
        compi_file = False  # same for component_index_classifications
        canopus_files = glob.glob(os.path.join(self._canopus_dir, "*"))
        possible_ci_file = [
            canopus_file
            for canopus_file in canopus_files
            if canopus_file.endswith("cluster_index_classifications.txt")
        ]
        possible_compi_file = [
            canopus_file
            for canopus_file in canopus_files
            if canopus_file.endswith("component_index_classifications.txt")
        ]
        if possible_ci_file:
            ci_file = possible_ci_file[0]
        if possible_compi_file:
            compi_file = possible_compi_file[0]

        if not os.path.isfile(ci_file):
            logger.info("Converting canopus output using canopus_treemap")
            try:
                analyse_canopus(self._canopus_dir, self._gnps_dir, self._canopus_dir)
            except FileNotFoundError as er1:
                logger.warning(f"{er1}. CANOPUS output is missing in {self._canopus_dir}")
            except Exception as er2:
                logger.warning(
                    f"canopus_treemap failed with: {er2}. Probably the MN "
                    f"version from GNPS is too old. Will attempt to read "
                    f"directly from canopus_dir"
                )
            else:
                # find processed canopus files again
                canopus_files = glob.glob(os.path.join(self._canopus_dir, "*"))
                possible_ci_file = [
                    canopus_file
                    for canopus_file in canopus_files
                    if canopus_file.endswith("cluster_index_classifications.txt")
                ]
                possible_compi_file = [
                    canopus_file
                    for canopus_file in canopus_files
                    if canopus_file.endswith("component_index_classifications.txt")
                ]
                if possible_ci_file:
                    ci_file = possible_ci_file[0]
                if possible_compi_file:
                    compi_file = possible_compi_file[0]

        if os.path.isfile(ci_file):
            spectra_classes_names, spectra_classes = self._read_spectra_classes(ci_file)

            if os.path.isfile(compi_file):
                mf_classes_names, mf_classes = self._read_mf_classes(compi_file)
                self._mf_classes = mf_classes
                self._mf_classes_names = mf_classes_names
                self._mf_classes_names_inds = {elem: i for i, elem in enumerate(mf_classes_names)}
        else:
            # use canopus output correctly (only for spectra)
            logger.info(
                "Attempting to read spectra classes directly from "
                "canopus_dir (canopus_summary.tsv)"
            )
            spectra_classes_names, spectra_classes = self._read_spectra_classes_directly()
            # mfs have to be added later with info about mf <- spectra
            # this happens with transfer_spec_classes_to_mfs() in loader.py

        self._spectra_classes = spectra_classes
        self._spectra_classes_names = spectra_classes_names
        self._spectra_classes_names_inds = {elem: i for i, elem in enumerate(spectra_classes_names)}

    def _read_spectra_classes(self, input_file):
        """Read canopus classes for spectra, return classes_names, classes.

        Args:
            input_file: str, cluster_index_classifications.txt
        Returns:
            Tuple of:
            - ci_classes_names: list of str - the names of each different level
            - ci_classes: dict of {str, lists of tuple(str, float)} - per spectrum index (key) the classes for each level
                where each level is a list of (class_name, score) sorted on best choice so index 0 is the best
                class prediction for a level. When no class is present, instead of Tuple it will be None for that level.
        """
        ci_classes = {}  # make a dict {ci: [[(class,score)]]}
        ci_classes_header = None
        ci_classes_names = []

        if os.path.isfile(input_file):
            logger.info(f"reading canopus results for spectra from {input_file}")
            with open(input_file) as inf:
                ci_classes_header = inf.readline().strip().split("\t")
                for line in inf:
                    line = line.strip("\n").split("\t")
                    classes_list = []
                    for lvl in line[3:]:
                        lvl_list = []
                        for l_class in lvl.split("; "):
                            if l_class:
                                l_class = l_class.split(":")
                                c_tup = tuple([l_class[0], float(l_class[1])])
                            else:
                                c_tup = None  # default value for class value
                            lvl_list.append(c_tup)
                        classes_list.append(lvl_list)
                    ci_classes[line[1]] = classes_list
        else:
            logger.warn(
                "could not load cluster_index_classifications.txt; missing from canopus_dir"
            )

        if ci_classes_header:
            #  todo: rename the output from the canopus script directly
            ci_classes_names = [f"cf_{elem}" for elem in ci_classes_header[3:-3]] + [
                f"npc_{elem}" for elem in ci_classes_header[-3:]
            ]
        return ci_classes_names, ci_classes

    def _read_spectra_classes_directly(self):
        """Read canopus classes directly from canopus output and save output.

        Returns:
            Tuple of:
            - can_classes_names: list of str - the names of each different level
            - can_classes: dict of {str, lists of tuple(str, float/None)} - per spectrum index (key) the classes for each level
                where each level is a list of (class_name, score) sorted on best choice so index 0 is the best
                class prediction for a level. When no class is present, instead of Tuple it will be None for that level.
        """
        can_classes = {}  # make a dict {spec: [[(class,score)]]}
        can_classes_header = None
        can_classes_names = []

        input_file = os.path.join(self._canopus_dir, "canopus_summary.tsv")
        sep = "\t"
        if os.path.isfile(input_file):
            logger.info(f"reading canopus results for spectra directly from " f"{input_file}")
            with open(input_file) as inf:
                can_classes_header = inf.readline().strip().split(sep)
                # get [4:8] - level 5 subclass        class   superclass
                starti = 4
                stopi = 8
                for line in inf:
                    line = line.strip("\n").split(sep)
                    # make tuples with scores as 1
                    classes_list = [[(cls, 1)] if cls else [None] for cls in line[starti:stopi]]
                    # assuming that all MGFs have ScanNumber headers..
                    spec_index = line[0].split("ScanNumber")[-1]
                    classes_list.reverse()  # reverse so superclass is first
                    can_classes[spec_index] = classes_list

            # use canopus_treemap to produce NPClassifier classes
            # TODO: probably change when sirius v5 comes out
            logger.info("Using canopus_treemap to get NPC classes")
            canopus_workspace = Canopus(sirius=self._canopus_dir)
            npc_file = os.path.join(self._canopus_dir, "npc_summary.tsv")
            canopus_workspace.npcSummary().to_csv(npc_file, sep=sep)
            with open(npc_file) as inf:
                npc_can_classes_header = inf.readline().strip().split(sep)
                # get [2:7] - name,directoryName,pathway,pathwayProbability,superclass,superclassProbability,class,classProbability,ClassyFirePrediction,ClassyFirePredictionProbability
                npc_starti = 2
                npc_stopi = 7
                for line in inf:
                    line = line.strip("\n").split(sep)
                    # make tuples with scores
                    npc_classes_list = [
                        [(cls, float(c_score))] if cls != "N/A" else [None]
                        for cls, c_score in zip(
                            line[npc_starti:npc_stopi:2], line[npc_starti + 1 : npc_stopi + 1 : 2]
                        )
                    ]
                    # assuming that all MGFs have ScanNumber headers..
                    spec_index = line[1].split("ScanNumber")[-1]
                    can_classes[spec_index].extend(npc_classes_list)
        else:
            logger.warn(
                f"could not load canopus results for spectra; "
                f"{input_file} missing from canopus_dir"
            )

        if can_classes_header:
            # important that the names here match with MIBiG
            can_classes_names = [f"cf_{elem}" for elem in can_classes_header[starti:stopi]]
            can_classes_names.reverse()  # reverse so superclass is first
            npc_can_classes_names = [
                f"npc_{elem}" for elem in npc_can_classes_header[npc_starti:npc_stopi:2]
            ]
            can_classes_names.extend(npc_can_classes_names)

            # make sure that non-existing NPC predictions are added as [None]
            # to all spectra - some spectra where not included in npc_summary
            len_classes_names = len(can_classes_names)
            for spec_index, classes_list in can_classes.items():
                diff = len_classes_names - len(classes_list)
                if diff > 0:
                    for _ in range(diff):
                        classes_list.append([None])

            # save output, next time clusterindex_classifications.txt is read
            # save just the info that we need in the format that is parsed in
            # read_spectra_classes()
            output_file = os.path.join(self._canopus_dir, "cluster_index_classifications.txt")
            with open(output_file, "w") as outf:
                header = ["componentindex", "cluster index", "formula"] + [
                    name.partition("_")[-1] for name in can_classes_names
                ]
                outf.write("\t".join(header) + "\n")
                for spec, classes in can_classes.items():
                    formatted_classes = []
                    for lvl in classes:
                        lvl_classes = []
                        for tup in lvl:
                            if not tup:
                                lvl_classes.append("")
                            else:
                                lvl_classes.append(f"{tup[0]}:{tup[1]:.3f}")
                        formatted_classes.append("; ".join(lvl_classes))
                    output_l = ["-", spec, "-"] + formatted_classes
                    outf.write("\t".join(output_l) + "\n")
        return can_classes_names, can_classes

    def _read_mf_classes(self, input_file):
        """Read canopus classes for mfs, return classes_names, classes.

        Args:
            input_file: str, component_index_classifications.txt
        Returns:
            Tuple of:
            - compi_classes_names: list of str - the names of each different level
            - compi_classes: dict of {str: lists of tuple(str, float)} - per mf index (key) the classes for each level
                where each level is a list of (class_name, fraction) sorted on best choice so index 0 is the best
                class prediction for a level. When no class is present, instead of Tuple it will be None for that level.
        """
        compi_classes = {}  # make a dict {compi: [[(class,score)]]}
        compi_classes_header = None
        compi_classes_names = []

        if os.path.isfile(input_file):
            with open(input_file) as inf:
                compi_classes_header = inf.readline().strip().split("\t")
                for line in inf:
                    line = line.strip("\n").split("\t")
                    classes_list = []
                    for lvl in line[2:]:
                        lvl_list = []
                        for l_class in lvl.split("; "):
                            if l_class:
                                l_class = l_class.split(":")
                                c_tup = tuple([l_class[0], float(l_class[1])])
                            else:
                                c_tup = None  # default value for class value
                            lvl_list.append(c_tup)
                        classes_list.append(lvl_list)
                    compi_classes[line[0]] = classes_list
        else:
            logger.warn(
                "could not load component_index_classifications.txt; missing from canopus_dir"
            )

        if compi_classes_header:
            #  todo: rename the output from the canopus script directly
            compi_classes_names = [f"cf_{elem}" for elem in compi_classes_header[2:-3]] + [
                f"npc_{elem}" for elem in compi_classes_header[-3:]
            ]
        return compi_classes_names, compi_classes

    def transfer_spec_classes_to_mfs(self, mfs, fraction_cutoff=0.0):
        """Set _mf_classes(_names) from spectra_classes and return classes.

        This can be used in the _loader to get mf classes when the GNPS MN
        version is too old and canopus_treemap fails to work directly.

        Args:
            mfs: list of MolecularFamily from the NPLinker space
            fraction_cutoff: float, cut-off for the fraction of class terms
                needed to be included in the mf
        Returns:
            dict of {str: lists of tuple(str, float)} - per mf (key) the classes for each level
                where each level is a list of (class_name, fraction) sorted on best choice so index 0 is the best
                class prediction for a level. When no class is present, instead of Tuple it will be None for that level.
        """
        self._mf_classes_names = self._spectra_classes_names
        self._mf_classes_names_inds = self._spectra_classes_names_inds
        mf_classes = {}

        for mf in mfs:
            fid = mf.id  # the key
            spectra = mf.spectra
            # if singleton family, format like 'fid_spectrum-id'
            if fid.startswith("singleton-"):
                spec_id = spectra[0].id
                fid += f"_{spec_id}"
            len_mf = len(spectra)

            classes_per_spectra = []
            for spec in spectra:
                spec_classes = self.spectra_classes.get(spec.id)
                if spec_classes:  # account for spectra without prediction
                    classes_per_spectra.append(spec_classes)

            if not classes_per_spectra:
                continue  # no spectra with classes for this mf

            sorted_classes = []
            for i, class_level in enumerate(self._mf_classes_names):
                # 1. aggregate classes from all spectra for this class level
                classes_cur_level = []
                for spec_classes in classes_per_spectra:
                    try:
                        for class_tup in spec_classes[i]:
                            if class_tup:
                                classes_cur_level.append(class_tup[0])
                    except IndexError:
                        print(self._mf_classes_names)
                        print(i, class_level)
                        print(classes_per_spectra)
                        print(spec_classes)
                        print(spectra)
                # 2. count the instances of each class in the MF per class lvl
                counts_cur_level = Counter(classes_cur_level)
                # 3. calculate fraction and sort high to low, filter out Nones
                fraction_tups = sorted(
                    (
                        (cls, count / len_mf)
                        for cls, count in counts_cur_level.most_common()
                        if count / len_mf >= fraction_cutoff
                    ),
                    key=lambda x: x[1],
                    reverse=True,
                )
                if not fraction_tups:
                    fraction_tups = [None]
                sorted_classes.append(fraction_tups)
            mf_classes[fid] = sorted_classes

        self._mf_classes = mf_classes
        return mf_classes

    def show(self, objects):
        """Show a table of predicted chemical compound classes for spectrum/MF.

        Args:
              objects: list of Spectrum or MolecularFamily objects
        Returns:
            pandas.DataFrame of objects (rows) vs classes (columns)
        """
        pass

    @property
    def spectra_classes(self):
        return self._spectra_classes

    @property
    def spectra_classes_names(self):
        return self._spectra_classes_names

    @property
    def spectra_classes_names_inds(self):
        return self._spectra_classes_names_inds

    @property
    def mf_classes(self):
        return self._mf_classes

    @property
    def mf_classes_names(self):
        return self._mf_classes_names

    @property
    def mf_classes_names_inds(self):
        return self._mf_classes_names_inds


class MolNetEnhancerResults:
    """Class for storing MolNetEnhancer results.

    The input file for ClassyFire results is read from the molnetenhancer directory:
        - ClassyFireResults_Network.txt
    """

    def __init__(self, mne_dir):
        """Read the class info from file in root_dir/molnetenhancer/.

        Args:
            mne_dir: str, mne_dir found in root_dir of nplinker project
        """
        cf_classes_names, mf_classes, spectra2mf = self._read_cf_classes(mne_dir)
        self._spectra2mf = spectra2mf
        self._mf_classes = mf_classes
        self._spectra_classes_names = cf_classes_names  # if NPC gets implemented, add here
        self._spectra_classes_names_inds = {elem: i for i, elem in enumerate(cf_classes_names)}

    def _read_cf_classes(self, mne_dir):
        r"""Read ClassyFireResults_Network.txt in molnetenhancer dir.

        Args:
            mne_dir: str, mne_dir found in root_dir of nplinker project
        Returns:
            tuple of:
            -list of str - names of the classes in order
            -dict of {str: [(str, float)]} - linking mfs to (classes, scores) in order of names,
                singleton families are denoted with S[\d]+
            -dict of {str:str} - linking spectra to mfs
        """
        columns = []
        mne_component_dict = {}
        mne_cluster2component = {}
        # flexible finding of CF results
        input_file = "not_found"
        wanted_file = "ClassyFireResults_Network.txt"
        possible_files = glob.glob(os.path.join(mne_dir, "*")) + glob.glob(
            os.path.join(mne_dir, "*", "*")
        )
        try:
            input_file = [
                pos_file for pos_file in possible_files if pos_file.endswith(wanted_file)
            ][0]
        except IndexError:
            pass

        if not os.path.isfile(input_file):
            logger.info(
                f"No MolNetEnhancer result present at {mne_dir}. "
                f"(run it on GNPS and download it here if you want to "
                f"use it)"
            )
            return columns, mne_component_dict, mne_cluster2component

        with open(input_file) as inf:
            logger.info(f"reading molnetenhancer results from {mne_dir}")
            header = inf.readline().strip().split("\t")
            # get the columns that are of interest to us
            columns = [
                "cf_direct_parent" if col == "CF_Dparent" else col.lower()
                for i, col in enumerate(header[3:])
                if i % 2 == 0
            ]
            for line in inf:
                line = line.strip("\n").split("\t")
                cluster = line.pop(0)
                component = line.pop(0)
                nr_nodes = line.pop(0)
                # todo: make it easier to query classes of singleton families
                # if singleton family, format like '-1_spectrum-id' like canopus results
                # Note that the singleton families id is "singleton-" + spectrum-id.
                if nr_nodes == "1":
                    component = f"-1_{cluster}"
                class_info = []
                # get class names and scores in (class, score)
                for i in range(0, len(line), 2):
                    cur_class = line[i]
                    if not cur_class or cur_class == "no matches":
                        # add None instead of tuple when no match at this lvl
                        class_tup = None
                        # catch when there is 'no matches' at superclass lvl ->
                        # no prediction and dont add to dict (same as canopus)
                        if columns[int(i / 2)] == "cf_kingdom":
                            break
                    else:
                        class_tup = (line[i], float(line[i + 1]))

                    class_info.append(class_tup)
                if component not in mne_component_dict and class_info:
                    mne_component_dict[component] = class_info
                mne_cluster2component[cluster] = component

        return columns, mne_component_dict, mne_cluster2component

    def spectra_classes(self, spectrum_id):
        """Return classes by relating spectrum_id in the mf_classes.

        Args:
            spectrum_id: int/str, spectrum_id - ints will be converted to str
        """
        classes = []
        if isinstance(spectrum_id, int):
            spectrum_id = str(spectrum_id)
        mf_id = self.spectra2mf.get(spectrum_id)
        if mf_id:
            classes = self.mf_classes.get(mf_id)
        return classes

    @property
    def spectra2mf(self):
        return self._spectra2mf

    @property
    def spectra_classes_names(self):
        return self._spectra_classes_names

    @property
    def spectra_classes_names_inds(self):
        return self._spectra_classes_names_inds

    @property
    def mf_classes(self):
        return self._mf_classes



from __future__ import annotations
import csv
from os import PathLike
from pathlib import Path
from nplinker.metabolomics.abc import AnnotationLoaderBase
from nplinker.utils import is_file_format


GNPS_UNIVERSAL_SPECTRUM_IDENTIFIER_URL = (
    "https://metabolomics-usi.gnps2.org/{}/?usi1=mzspec:GNPS:GNPS-LIBRARY:accession:{}"
)


class GNPSAnnotationLoader(AnnotationLoaderBase):
    """Load annotations from GNPS output file.

    ??? info "Concept"
        [GNPS data][gnps-data]

    The annotation file is a `.tsv` file from GNPS output archive, as described
    below for each GNPS workflow type:

    1. METABOLOMICS-SNETS
        - result_specnets_DB/*.tsv
    2. METABOLOMICS-SNETS-V2
        - result_specnets_DB/.tsv
    3. FEATURE-BASED-MOLECULAR-NETWORKING
        - DB_result/*.tsv
    """

    def __init__(self, file: str | PathLike) -> None:
        """Initialize the GNPSAnnotationLoader.

        Args:
            file: The GNPS annotation file.

        Examples:
            >>> loader = GNPSAnnotationLoader("gnps_annotations.tsv")
            >>> print(loader.annotations["100"])
            {'#Scan#': '100',
            'Adduct': 'M+H',
            'CAS_Number': 'N/A',
            'Charge': '1',
            'Compound_Name': 'MLS002153841-01!Iobenguane sulfate',
            'Compound_Source': 'NIH Pharmacologically Active Library',
            'Data_Collector': 'VP/LMS',
            'ExactMass': '274.992',
            'INCHI': 'N/A',
            'INCHI_AUX': 'N/A',
            'Instrument': 'qTof',
            'IonMode': 'Positive',
            'Ion_Source': 'LC-ESI',
            'LibMZ': '276.003',
            'LibraryName': 'lib-00014.mgf',
            'LibraryQualityString': 'Gold',
            'Library_Class': '1',
            'MQScore': '0.704152',
            'MZErrorPPM': '405416',
            'MassDiff': '111.896',
            'Organism': 'GNPS-NIH-SMALLMOLECULEPHARMACOLOGICALLYACTIVE',
            'PI': 'Dorrestein',
            'Precursor_MZ': '276.003',
            'Pubmed_ID': 'N/A',
            'RT_Query': '795.979',
            'SharedPeaks': '7',
            'Smiles': 'NC(=N)NCc1cccc(I)c1.OS(=O)(=O)O',
            'SpecCharge': '1',
            'SpecMZ': '164.107',
            'SpectrumFile': 'spectra/specs_ms.pklbin',
            'SpectrumID': 'CCMSLIB00000086167',
            'TIC_Query': '986.997',
            'UpdateWorkflowName': 'UPDATE-SINGLE-ANNOTATED-GOLD',
            'tags': ' ',
            'png_url': 'https://metabolomics-usi.gnps2.org/png/?usi1=mzspec:GNPS:GNPS-LIBRARY:accession:CCMSLIB00000086167',
            'json_url': 'https://metabolomics-usi.gnps2.org/json/?usi1=mzspec:GNPS:GNPS-LIBRARY:accession:CCMSLIB00000086167',
            'svg_url': 'https://metabolomics-usi.gnps2.org/svg/?usi1=mzspec:GNPS:GNPS-LIBRARY:accession:CCMSLIB00000086167',
            'spectrum_url': 'https://metabolomics-usi.gnps2.org/spectrum/?usi1=mzspec:GNPS:GNPS-LIBRARY:accession:CCMSLIB00000086167'}
        """
        self._file = Path(file)
        self._annotations: dict[str, dict] = {}

        self._validate()
        self._load()

    @property
    def annotations(self) -> dict[str, dict]:
        """Get annotations.

        Returns:
            Keys are spectrum ids ("#Scan#" in annotation file) and values are the annotations dict
            for each spectrum.
        """
        return self._annotations

    def _validate(self) -> None:
        """Validate the annotation file.

        Raises:
            ValueError: Raises ValueError if the file is not valid.
        """
        # validate file format
        if not is_file_format(self._file, "tsv"):
            raise ValueError(
                f"Invalid GNPS annotation file '{self._file}'. " f"Expected a .tsv file."
            )

        # validate required columns against the header
        required_columns = ["#Scan#", "Compound_Name", "Organism", "MQScore", "SpectrumID"]
        with open(self._file, mode="rt") as f:
            header = f.readline()
            for k in required_columns:
                if k not in header:
                    raise ValueError(
                        f"Invalid GNPS annotation file '{self._file}'. "
                        f"Expected a header line with '{k}' column, "
                        f"but got '{header}'."
                    )

        # validate that "#Scan#" must be unique
        with open(self._file, mode="rt") as f:
            reader = csv.DictReader(f, delimiter="\t")
            scans = [row["#Scan#"] for row in reader]
        duplicates = {x for x in scans if scans.count(x) > 1}
        if len(duplicates) > 0:
            raise ValueError(
                f"Invalid GNPS annotation file '{self._file}'. "
                f"Expected unique '#Scan#', but found duplicates '{duplicates}'."
            )

    def _load(self) -> None:
        """Load the annotations from the file."""
        with open(self._file, mode="rt") as f:
            dict_reader = csv.DictReader(f, delimiter="\t")
            for row in dict_reader:
                scan_id = row["#Scan#"]
                self._annotations[scan_id] = row
                # insert useful URLs
                for t in ["png", "json", "svg", "spectrum"]:
                    self._annotations[scan_id][f"{t}_url"] = (
                        GNPS_UNIVERSAL_SPECTRUM_IDENTIFIER_URL.format(t, row["SpectrumID"])
                    )



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle
import numpy
import scipy.io
from cdk_pywrapper.cdk_pywrapper import Compound


def main(datapath):
    data = scipy.io.loadmat(datapath + "/data_GNPS.mat")

    fp_size = 1024

    new_fps = []

    for i in data["inchi"]:
        print("Processing sample %s" % i)
        inchi = i[0][0]
        print(inchi)
        c = Compound(compound_string=inchi, identifier_type="inchi")
        fp = c.get_fingerprint()
        fp_array = numpy.zeros(fp_size)
        for fp_bit in range(fp_size):
            fp_array[fp_bit] = fp.get(fp_bit)
        new_fps.append(fp_array)

    with open(datapath + "/cdk_fingerprints.bin", "wb") as f:
        pickle.dump(new_fps, f)


def default_fingerprint_from_inchi(inchi):
    c = Compound(compound_string=inchi, identifier_type="inchi")
    # this should accommodate different types!
    fp = c.get_fingerprint()
    fp_size = 1024
    fp_array = numpy.zeros(fp_size)
    for fp_bit in range(fp_size):
        fp_array[fp_bit] = fp.get(fp_bit)
    return fp_array


def fingerprint_from_smiles(smiles, fingerprint_type=None):
    if fingerprint_type is None:
        fingerprint = numpy.array([])
        for fp_type in ("cdk_default", "substructure", "klekota-roth"):
            fingerprint = numpy.hstack((fingerprint, fingerprint_from_smiles(smiles, fp_type)))
        return fingerprint

    c = Compound(compound_string=smiles, identifier_type="smiles")

    if fingerprint_type == "cdk_default":
        fingerprinter = c.cdk.fingerprint.Fingerprinter()
    elif fingerprint_type == "substructure":
        fingerprinter = c.cdk.fingerprint.SubstructureFingerprinter()
    elif fingerprint_type == "klekota-roth":
        fingerprinter = c.cdk.fingerprint.KlekotaRothFingerprinter()
    else:
        raise SystemExit(f"Unknown fingerprint type: {fingerprint_type}")

    fp = fingerprinter.getBitFingerprint(c.mol_container)
    fp_size = fp.size()
    fp_array = numpy.zeros(fp_size)
    for fp_bit in range(fp_size):
        fp_array[fp_bit] = fp.get(fp_bit)
    return fp_array


def fingerprint_from_inchi(inchi, fingerprint_type=None):
    if fingerprint_type is None:
        fingerprint = numpy.array([])
        for fp_type in ("cdk_default", "substructure", "klekota-roth"):
            fingerprint = numpy.hstack((fingerprint, fingerprint_from_inchi(inchi, fp_type)))
        return fingerprint

    c = Compound(compound_string=inchi, identifier_type="inchi")

    if fingerprint_type == "cdk_default":
        fingerprinter = c.cdk.fingerprint.Fingerprinter()
    elif fingerprint_type == "substructure":
        fingerprinter = c.cdk.fingerprint.SubstructureFingerprinter()
    elif fingerprint_type == "klekota-roth":
        fingerprinter = c.cdk.fingerprint.KlekotaRothFingerprinter()
    else:
        raise SystemExit(f"Unknown fingerprint type: {fingerprint_type}")

    fp = fingerprinter.getBitFingerprint(c.mol_container)
    fp_size = fp.size()
    fp_array = numpy.zeros(fp_size)
    for fp_bit in range(fp_size):
        fp_array[fp_bit] = fp.get(fp_bit)
    return fp_array



# CG: This module parse AntiSMASH files to get the predicted monomers of product.
# And it only cares the case when the monomers are amino acids.
# TODO This module is used in bgc.py, referenced in gcf.py and then in misc.py,
# NPLinker core business does not use it at all.

from __future__ import annotations
import argparse
from Bio import SeqIO


AA_CODES = [
    "ala",
    "cys",
    "asp",
    "glu",
    "phe",
    "gly",
    "his",
    "ile",
    "lys",
    "leu",
    "met",
    "asn",
    "pyl",
    "pro",
    "gln",
    "arg",
    "ser",
    "thr",
    "sec",
    "val",
    "trp",
    "tyr",
]

# AA_CODES_ISOMER contains codes that should map to
# an entry in the AA_CODES list - i.e. isomers, etc.
AA_CODES_ISOMER = {}
for c in AA_CODES:
    AA_CODES_ISOMER["d-%s" % c] = c
AA_CODES_ISOMER["b-ala"] = "ala"


# CG: create one class for handling antismash output files and records
## since antismash output formats change along with version
## we should delegate all processes in other classes/functions to the methods of this class
## so that one type of data, one class covering all methodes needed
class AntiSmashFile:
    def __init__(self, filename):
        """Parse product and specificity (predicted monomer) from AntiSMASH file.

        This class is a wrapper for AntiSmash4Record and AntiSmash5Record.

        Deprecated:
            AntiSMASH file always contains one sequence record, so users should
            directly use AntiSmash5Record or AntiSmash4Record.

        Args:
            filename: AntiSMASH file path
        """
        self.raw_data = []
        self.filename = filename

        with open(filename) as f:
            # antiSMASH 5 vs. 4 output is vastly different
            # CG: TODO should use SeqIO.read since antismash file has only one sequence record
            for record in SeqIO.parse(f, "gb"):
                if "structured_comment" in record.annotations:
                    as_version = record.annotations["structured_comment"]["antiSMASH-Data"][
                        "Version"
                    ]
                    as_major_version = as_version.split(".")[0]
                    if as_major_version == "5":
                        self.raw_data.append(AntiSmash5Record(record))
                    else:
                        raise ValueError(f"Invalid antiSMASH version: {as_version}")
                else:
                    self.raw_data.append(AntiSmash4Record(record))

    def get_spec(self):
        """Get specificity (predicted monomer)."""
        for r in self.raw_data:
            yield from r.get_spec()

    @property
    def products(self):
        """Get list of products."""
        return [x.products for x in self.raw_data]

    def build_prob(self):
        [x.build_prob() for x in self.raw_data]

    def get_prob(self, aa):
        return [x.get_prob(aa) for x in self.raw_data]


class AntiSmash5Record:
    # CG: TODO input should also include antismash file
    def __init__(self, seq_record):
        """Parse product and specificity (predicted monomer of product) from
            AntiSMASH v5 data.

        Args:
            seq_record: SeqRecord of AntiSMASH
        """
        self.raw_data = seq_record
        self.description = self.raw_data.description

        # parse "cand_cluster"
        clusters = [x for x in self.raw_data.features if x.type == "cand_cluster"]

        products = []
        for prod_list in [x.qualifiers["product"] for x in clusters]:
            for prod in prod_list:
                products.append(prod)

        smiles = []
        for smiles_list in [x.qualifiers["SMILES"] for x in clusters if "SMILES" in x.qualifiers]:
            for s in smiles_list:
                smiles.append(s)

        # parse "aSDomain"
        asdomains = [x for x in self.raw_data.features if x.type == "aSDomain"]
        asdomains_with_predictions = [x for x in asdomains if "specificity" in x.qualifiers]
        asdomains_with_predictions_known = [
            x for x in asdomains_with_predictions if "AMP-binding" in x.qualifiers["aSDomain"]
        ]

        self.products = products
        self.smiles = smiles
        self.asdomains_with_predictions_known = asdomains_with_predictions_known

        # parse predicted monomer of product, i.e. "specificity"
        self.specificities = []
        specificities_listed = [
            x.qualifiers["specificity"] for x in asdomains_with_predictions_known
        ]
        for specificities_single_list in specificities_listed:
            for specificity in specificities_single_list:
                if specificity.startswith("consensus: "):
                    self.specificities.append(specificity.split()[1])

    def get_prob(self, aa):
        """Get probability of predicted amino acid.

        Args:
            aa: amino acid

        Returns:
            prediction probability
        """
        if aa in self.specificities:
            return 1.0
        else:
            return 0.0

    def get_spec(self):
        """Get predicted specificities (animo acids)."""
        for aa in set(self.specificities):
            if aa in AA_CODES:
                yield aa
            elif aa in AA_CODES_ISOMER:
                yield AA_CODES_ISOMER[aa]


class AntiSmash4Record:
    # CG: this is for antiSMASH version 4.0 results, it can be removed.

    def __init__(self, seq_record):
        self.raw_data = seq_record
        self.description = self.raw_data.description

        clusters = [x for x in self.raw_data.features if x.type == "cluster"]

        products = [x.qualifiers["product"] for x in clusters]
        smiles = [
            [y for y in x.qualifiers["note"] if y.startswith("Monomers prediction")]
            for x in clusters
        ]

        asdomains = [x for x in self.raw_data.features if x.type == "aSDomain"]
        asdomains_with_predictions = [x for x in asdomains if "specificity" in x.qualifiers]
        asdomains_with_predictions_known = [
            x for x in asdomains_with_predictions if "AMP-binding" in x.qualifiers["domain"]
        ]

        self.products = products
        self.smiles = smiles
        self.asdomains_with_predictions_known = asdomains_with_predictions_known

        self.build_prob()

    def display(self):
        print(self.products)
        print(self.smiles)
        print(self.asdomains_with_predictions_known)

        print(self.asdomains_with_predictions_known[0].qualifiers["specificity"])

    def get_probable_aa(self):
        aalist = []
        for domain in self.specificities:
            for vote in domain:
                aalist.extend(vote)
        aaset = set(aalist)
        if "none" in aaset:
            aaset.remove("none")
        return aaset

    def get_spec(self):
        for domain in self.asdomains_with_predictions_known:
            yield process_specificity(domain.qualifiers["specificity"])
            # for x in domain.qualifiers['specificity']:
            #     if x.startswith('PID to NN:') or x.startswith('SNN score:'):
            #         continue
            #     else:
            #         yield x

            # x_split = x.split(': ')
            # if len(x_split) > 1:
            #    yield x_split[1]
            # else:
            #    x_split = x.split(' ')
            #    yield x_split[1]

    def build_prob(self):
        self.specificities = [
            process_specificity(x.qualifiers["specificity"])
            for x in self.asdomains_with_predictions_known
        ]

    def get_prob(self, aa):
        # 'none' votes do not affect predictions, because then
        # all AAs are equally likely, so we can't use them to exclude.
        prob = 1.0
        for domain in self.specificities:
            domain_vote_sum = 0
            for votes in domain:
                for vote in set(votes):
                    if vote == "none":
                        p_vote = 0
                    #     p_vote = 1.0 / len(AA_CODES)
                    elif vote == aa:
                        p_vote = 1.0 / len(set(votes))
                    else:
                        p_vote = 0
                    domain_vote_sum += p_vote
            domain_prob = domain_vote_sum / len(domain)
            prob *= 1 - domain_prob

        return 1 - prob


def stachelhaus(prediction_string):
    prediction_string = prediction_string.split(":")[-1].strip()
    predictions = predict_aa_string(prediction_string)
    return predictions


def sandpuma(prediction_string):
    prediction_string = prediction_string.split(":")[-1].strip()
    predictions = predict_aa_string(prediction_string)
    return predictions


def predicat(prediction_string):
    prediction_string = prediction_string.split("-")[-1]
    predictions = predict_aa_string(prediction_string)
    return predictions


def phmm(prediction_string):
    prediction_string = prediction_string.split(":")[-1].strip()
    predictions = predict_aa_string(prediction_string)
    return predictions


def nrpspredictor3(prediction_string):
    prediction_string = prediction_string.split(":")[-1].strip()
    predictions = predict_aa_string(prediction_string)
    return predictions


def predict_aa_string(prediction_string):
    parsed_predictions = []
    for prediction in prediction_string.split("|"):
        if prediction in AA_CODES:
            parsed_predictions.append(prediction)
        elif prediction in AA_CODES_ISOMER:
            parsed_predictions.append(AA_CODES_ISOMER[prediction])
        else:
            parsed_predictions.append("none")
    return parsed_predictions


def process_specificity(prediction_list):
    votes = []
    for prediction_string in prediction_list:
        if prediction_string.startswith("PID to NN:") or prediction_string.startswith("SNN score:"):
            continue
        if prediction_string.startswith("SANDPUMA"):
            prediction = sandpuma(prediction_string)
        else:
            continue
        # Only process SANDPUMA votes for now (to not have to collate the votes)
        # if prediction_string.startswith('Stachelhaus code:'):
        #     prediction = stachelhaus(prediction_string)
        # elif prediction_string.startswith('NRPSpredictor3 SVM'):
        #     prediction = nrpspredictor3(prediction_string)
        # elif prediction_string.startswith('pHMM'):
        #     prediction = phmm(prediction_string)
        # elif prediction_string.startswith('PrediCAT'):
        #     prediction = predicat(prediction_string)
        # elif prediction_string.startswith('SANDPUMA'):
        #     prediction = sandpuma(prediction_string)
        # else:
        #     prediction = []
        votes.append(prediction)
    return votes


def to_set(nested):
    for item in nested:
        if type(item) is str:
            yield item
        else:
            yield from to_set(item)


def predict_aa(filename):
    asr = AntiSmashFile(filename)
    for item in asr.raw_data:
        # for aa in item.get_probable_aa():
        for aa in AA_CODES:
            res = item.get_prob(aa)
            yield aa, res


def read_aa_losses(filename):
    """Read AA losses from data file. (assume fixed structure...)."""
    aa_losses = {}
    with open(filename) as f:
        reader = csv.reader(f, delimiter=",")
        next(reader)  # skip headers
        for line in reader:
            if len(line) == 0:
                continue
            aa_id = line[1]
            aa_mono = float(line[4])
            aa_avg = float(line[5])
            aa_losses[aa_id.lower()] = (aa_mono, aa_avg)

    return aa_losses


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Predict AA specificity for gbk file")
    parser.add_argument("file", help=".gbk file with specificity predictions")
    args = parser.parse_args()

    for aa, res in predict_aa(args.file):
        print(f"{aa},{res}")



from .gnps_annotation_loader import GNPSAnnotationLoader
from .gnps_downloader import GNPSDownloader
from .gnps_extractor import GNPSExtractor
from .gnps_file_mapping_loader import GNPSFileMappingLoader
from .gnps_format import GNPSFormat
from .gnps_format import gnps_format_from_archive
from .gnps_format import gnps_format_from_file_mapping
from .gnps_format import gnps_format_from_task_id
from .gnps_molecular_family_loader import GNPSMolecularFamilyLoader
from .gnps_spectrum_loader import GNPSSpectrumLoader


__all__ = [
    "GNPSAnnotationLoader",
    "GNPSDownloader",
    "GNPSExtractor",
    "GNPSFileMappingLoader",
    "GNPSFormat",
    "GNPSMolecularFamilyLoader",
    "GNPSSpectrumLoader",
    "gnps_format_from_archive",
    "gnps_format_from_file_mapping",
    "gnps_format_from_task_id",
]



class RosettaHit:
    def __init__(self, spec, gnps_id, mibig_id, bgc, spec_match_score, bgc_match_score):
        self.spec = spec
        self.gnps_id = gnps_id
        self.mibig_id = mibig_id
        self.bgc = bgc
        self.spec_match_score = spec_match_score
        self.bgc_match_score = bgc_match_score

    def __str__(self):
        return "RosettaHit: {}<-->{} via ({} ({:.3f}), {} ({:.3f}))".format(
            self.spec.id,
            self.bgc.id,
            self.gnps_id,
            self.spec_match_score,
            self.mibig_id,
            self.bgc_match_score,
        )

    def __repr__(self):
        return str(self)



from .link_graph import LinkGraph
from .metcalf_scoring import MetcalfScoring
from .score import Score
from .scoring_method import ScoringMethod


__all__ = [
    "LinkGraph",
    "MetcalfScoring",
    "Score",
    "ScoringMethod",
]



from __future__ import annotations
import json
import logging
from collections.abc import Mapping
from collections.abc import Sequence
from os import PathLike
from pathlib import Path
from jsonschema import validate
from nplinker.defaults import GENOME_BGC_MAPPINGS_FILENAME
from nplinker.schemas import GENOME_BGC_MAPPINGS_SCHEMA
from nplinker.schemas import validate_podp_json
from nplinker.strain import StrainCollection
from nplinker.utils import list_dirs
from nplinker.utils import list_files
from ..genomics.antismash.podp_antismash_downloader import GenomeStatus
from ..genomics.antismash.podp_antismash_downloader import get_best_available_genome_id
from .bgc import BGC
from .gcf import GCF


logger = logging.getLogger(__name__)


def generate_mappings_genome_id_bgc_id(
    bgc_dir: str | PathLike, output_file: str | PathLike | None = None
) -> None:
    """Generate a file that maps genome id to BGC id.

    The input `bgc_dir` must follow the structure of the `antismash` directory defined in
    [Working Directory Structure][working-directory-structure], e.g.:
    ```shell
    bgc_dir
        ├── genome_id_1
        │  ├── bgc_id_1.gbk
        │  └── ...
        ├── genome_id_2
        │  ├── bgc_id_2.gbk
        │  └── ...
        └── ...
    ```

    Args:
        bgc_dir: The directory has one-layer of subfolders and each subfolder contains BGC files
            in `.gbk` format.

            It assumes that

            - the subfolder name is the genome id (e.g. refseq),
            - the BGC file name is the BGC id.
        output_file: The path to the output file.
            The file will be overwritten if it already exists.

            Defaults to None, in which case the output file will be placed in
            the directory `bgc_dir` with the file name
            [GENOME_BGC_MAPPINGS_FILENAME][nplinker.defaults.GENOME_BGC_MAPPINGS_FILENAME].
    """
    bgc_dir = Path(bgc_dir)
    genome_bgc_mappings = {}

    for subdir in list_dirs(bgc_dir):
        genome_id = Path(subdir).name
        bgc_files = list_files(subdir, suffix=(".gbk"), keep_parent=False)
        bgc_ids = [bgc_id for f in bgc_files if (bgc_id := Path(f).stem) != genome_id]
        if bgc_ids:
            genome_bgc_mappings[genome_id] = bgc_ids
        else:
            logger.warning("No BGC files found in %s", subdir)

    # sort mappings by genome_id and construct json data
    genome_bgc_mappings = dict(sorted(genome_bgc_mappings.items()))
    json_data_mappings = [{"genome_ID": k, "BGC_ID": v} for k, v in genome_bgc_mappings.items()]
    json_data = {"mappings": json_data_mappings, "version": "1.0"}

    # validate json data
    validate(instance=json_data, schema=GENOME_BGC_MAPPINGS_SCHEMA)

    if output_file is None:
        output_file = bgc_dir / GENOME_BGC_MAPPINGS_FILENAME
    with open(output_file, "w") as f:
        json.dump(json_data, f)
    logger.info("Generated genome-BGC mappings file: %s", output_file)


def add_strain_to_bgc(
    strains: StrainCollection, bgcs: Sequence[BGC]
) -> tuple[list[BGC], list[BGC]]:
    """Assign a Strain object to `BGC.strain` for input BGCs.

    BGC id is used to find the corresponding Strain object. It's possible that
    no Strain object is found for a BGC id.

    !!! Note
        The input `bgcs` will be changed in place.

    Args:
        strains: A collection of all strain objects.
        bgcs: A list of BGC objects.

    Returns:
        A tuple of two lists of BGC objects,

            - the first list contains BGC objects that are updated with Strain object;
            - the second list contains BGC objects that are not updated with
                Strain object because no Strain object is found.

    Raises:
        ValueError: Multiple strain objects found for a BGC id.
    """
    bgc_with_strain = []
    bgc_without_strain = []
    for bgc in bgcs:
        try:
            strain_list = strains.lookup(bgc.id)
        except ValueError:
            bgc_without_strain.append(bgc)
            continue
        if len(strain_list) > 1:
            raise ValueError(
                f"Multiple strain objects found for BGC id '{bgc.id}'."
                f"BGC object accept only one strain."
            )
        bgc.strain = strain_list[0]
        bgc_with_strain.append(bgc)

    logger.info(
        f"{len(bgc_with_strain)} BGC objects updated with Strain object.\n"
        f"{len(bgc_without_strain)} BGC objects not updated with Strain object."
    )
    return bgc_with_strain, bgc_without_strain


def add_bgc_to_gcf(
    bgcs: Sequence[BGC], gcfs: Sequence[GCF]
) -> tuple[list[GCF], list[GCF], dict[GCF, set[str]]]:
    """Add BGC objects to GCF object based on GCF's BGC ids.

    The attribute of `GCF.bgc_ids` contains the ids of BGC objects. These ids
    are used to find BGC objects from the input `bgcs` list. The found BGC
    objects are added to the `bgcs` attribute of GCF object. It is possible that
    some BGC ids are not found in the input `bgcs` list, and so their BGC
    objects are missing in the GCF object.

    !!! note
        This method changes the lists `bgcs` and `gcfs` in place.

    Args:
        bgcs: A list of BGC objects.
        gcfs: A list of GCF objects.

    Returns:
        A tuple of two lists and a dictionary,

            - The first list contains GCF objects that are updated with BGC objects;
            - The second list contains GCF objects that are not updated with BGC objects
                because no BGC objects are found;
            - The dictionary contains GCF objects as keys and a set of ids of missing
                BGC objects as values.
    """
    bgc_dict = {bgc.id: bgc for bgc in bgcs}
    gcf_with_bgc = []
    gcf_without_bgc = []
    gcf_missing_bgc: dict[GCF, set[str]] = {}
    for gcf in gcfs:
        for bgc_id in gcf.bgc_ids:
            try:
                bgc = bgc_dict[bgc_id]
            except KeyError:
                if gcf not in gcf_missing_bgc:
                    gcf_missing_bgc[gcf] = {bgc_id}
                else:
                    gcf_missing_bgc[gcf].add(bgc_id)
                continue
            gcf.add_bgc(bgc)

        if gcf.bgcs:
            gcf_with_bgc.append(gcf)
        else:
            gcf_without_bgc.append(gcf)

    logger.info(
        f"{len(gcf_with_bgc)} GCF objects updated with BGC objects.\n"
        f"{len(gcf_without_bgc)} GCF objects not updated with BGC objects.\n"
        f"{len(gcf_missing_bgc)} GCF objects have missing BGC objects."
    )
    return gcf_with_bgc, gcf_without_bgc, gcf_missing_bgc


def get_mibig_from_gcf(gcfs: Sequence[GCF]) -> tuple[list[BGC], StrainCollection]:
    """Get MIBiG BGCs and strains from GCF objects.

    Args:
        gcfs: A list of GCF objects.

    Returns:
        A tuple of two objects,

            - the first is a list of MIBiG BGC objects used in the GCFs;
            - the second is a StrainCollection object that contains all Strain objects used in the
            GCFs.
    """
    mibig_bgcs_in_use = []
    mibig_strains_in_use = StrainCollection()
    for gcf in gcfs:
        for bgc in gcf.bgcs:
            if bgc.is_mibig():
                mibig_bgcs_in_use.append(bgc)
                if bgc.strain is not None:
                    mibig_strains_in_use.add(bgc.strain)
    return mibig_bgcs_in_use, mibig_strains_in_use


# ------------------------------------------------------------------------------
# Functions to extract mappings for genomics side:
# strain_id <-> original_geonme_id <-> resolved_genome_id <-> bgc_id
# ------------------------------------------------------------------------------
def extract_mappings_strain_id_original_genome_id(
    podp_project_json_file: str | PathLike,
) -> dict[str, set[str]]:
    """Extract mappings "strain_id <-> original_genome_id".

    !!! tip
        The `podp_project_json_file` is the JSON file downloaded from PODP platform.

        For example, for PODP project MSV000079284, its JSON file is
        https://pairedomicsdata.bioinformatics.nl/api/projects/4b29ddc3-26d0-40d7-80c5-44fb6631dbf9.4.

    Args:
        podp_project_json_file: The path to the PODP project
            JSON file.

    Returns:
        Key is strain id and value is a set of original genome ids.

    See Also:
        - [podp_generate_strain_mappings][nplinker.strain.utils.podp_generate_strain_mappings]:
            Generate strain mappings JSON file for PODP pipeline.
    """
    mappings_dict: dict[str, set[str]] = {}
    with open(podp_project_json_file, "r") as f:
        json_data = json.load(f)

    validate_podp_json(json_data)

    for record in json_data["genomes"]:
        strain_id = record["genome_label"]
        genome_id = get_best_available_genome_id(record["genome_ID"])
        if genome_id is None:
            logger.warning("Failed to extract genome ID from genome with label %s", strain_id)
            continue
        if strain_id in mappings_dict:
            mappings_dict[strain_id].add(genome_id)
        else:
            mappings_dict[strain_id] = {genome_id}
    return mappings_dict


def extract_mappings_original_genome_id_resolved_genome_id(
    genome_status_json_file: str | PathLike,
) -> dict[str, str]:
    """Extract mappings "original_genome_id <-> resolved_genome_id".

    !!! tip
        The `genome_status_json_file` is generated by the [podp_download_and_extract_antismash_data]
        [nplinker.genomics.antismash.podp_antismash_downloader.podp_download_and_extract_antismash_data]
        function with a default file name [GENOME_STATUS_FILENAME][nplinker.defaults.GENOME_STATUS_FILENAME].

    Args:
        genome_status_json_file: The path to the genome status JSON file.


    Returns:
        Key is original genome id and value is resolved genome id.

    See Also:
        - [podp_generate_strain_mappings][nplinker.strain.utils.podp_generate_strain_mappings]:
            Generate strain mappings JSON file for PODP pipeline.
    """
    gs_mappings_dict = GenomeStatus.read_json(genome_status_json_file)
    return {gs.original_id: gs.resolved_refseq_id for gs in gs_mappings_dict.values()}


def extract_mappings_resolved_genome_id_bgc_id(
    genome_bgc_mappings_file: str | PathLike,
) -> dict[str, set[str]]:
    """Extract mappings "resolved_genome_id <-> bgc_id".

    !!! tip
        The `genome_bgc_mappings_file` is usually generated by the
        [generate_mappings_genome_id_bgc_id][nplinker.genomics.utils.generate_mappings_genome_id_bgc_id]
        function with a default file name [GENOME_BGC_MAPPINGS_FILENAME][nplinker.defaults.GENOME_BGC_MAPPINGS_FILENAME].

    Args:
        genome_bgc_mappings_file: The path to the genome BGC
            mappings JSON file.

    Returns:
        Key is resolved genome id and value is a set of BGC ids.

    See Also:
        - [podp_generate_strain_mappings][nplinker.strain.utils.podp_generate_strain_mappings]:
            Generate strain mappings JSON file for PODP pipeline.
    """
    with open(genome_bgc_mappings_file, "r") as f:
        json_data = json.load(f)

    # validate the JSON data
    validate(json_data, GENOME_BGC_MAPPINGS_SCHEMA)

    return {mapping["genome_ID"]: set(mapping["BGC_ID"]) for mapping in json_data["mappings"]}


def get_mappings_strain_id_bgc_id(
    mappings_strain_id_original_genome_id: Mapping[str, set[str]],
    mappings_original_genome_id_resolved_genome_id: Mapping[str, str],
    mappings_resolved_genome_id_bgc_id: Mapping[str, set[str]],
) -> dict[str, set[str]]:
    """Get mappings "strain_id <-> bgc_id".

    Args:
        mappings_strain_id_original_genome_id: Mappings "strain_id <-> original_genome_id".
        mappings_original_genome_id_resolved_genome_id: Mappings "original_genome_id <-> resolved_genome_id".
        mappings_resolved_genome_id_bgc_id: Mappings "resolved_genome_id <-> bgc_id".

    Returns:
        Key is strain id and value is a set of BGC ids.

    See Also:
        - `extract_mappings_strain_id_original_genome_id`: Extract mappings
            "strain_id <-> original_genome_id".
        - `extract_mappings_original_genome_id_resolved_genome_id`: Extract mappings
            "original_genome_id <-> resolved_genome_id".
        - `extract_mappings_resolved_genome_id_bgc_id`: Extract mappings
            "resolved_genome_id <-> bgc_id".
        - [podp_generate_strain_mappings][nplinker.strain.utils.podp_generate_strain_mappings]:
            Generate strain mappings JSON file for PODP pipeline.
    """
    mappings_dict = {}
    for strain_id, original_genome_ids in mappings_strain_id_original_genome_id.items():
        bgc_ids = set()
        for original_genome_id in original_genome_ids:
            resolved_genome_id = mappings_original_genome_id_resolved_genome_id[original_genome_id]
            if (bgc_id := mappings_resolved_genome_id_bgc_id.get(resolved_genome_id)) is not None:
                bgc_ids.update(bgc_id)
        if bgc_ids:
            mappings_dict[strain_id] = bgc_ids
    return mappings_dict



from pathlib import Path
from typing import Final


# The path to the NPLinker application database directory
NPLINKER_APP_DATA_DIR: Final = Path(__file__).parent / "data"


STRAIN_MAPPINGS_FILENAME: Final = "strain_mappings.json"
GENOME_BGC_MAPPINGS_FILENAME: Final = "genome_bgc_mappings.json"
GENOME_STATUS_FILENAME: Final = "genome_status.json"
GNPS_SPECTRA_FILENAME: Final = "spectra.mgf"
GNPS_MOLECULAR_FAMILY_FILENAME: Final = "molecular_families.tsv"
GNPS_ANNOTATIONS_FILENAME: Final = "annotations.tsv"
GNPS_FILE_MAPPINGS_TSV: Final = "file_mappings.tsv"
GNPS_FILE_MAPPINGS_CSV: Final = "file_mappings.csv"
STRAINS_SELECTED_FILENAME: Final = "strains_selected.json"


DOWNLOADS_DIRNAME: Final = "downloads"
MIBIG_DIRNAME: Final = "mibig"
GNPS_DIRNAME: Final = "gnps"
ANTISMASH_DIRNAME: Final = "antismash"
BIGSCAPE_DIRNAME: Final = "bigscape"
BIGSCAPE_RUNNING_OUTPUT_DIRNAME: Final = "bigscape_running_output"
OUTPUT_DIRNAME: Final = "output"



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import sys
import time
import numpy
import scipy.io
from . import mk_fprints
from . import spectrum
from . import spectrum_filters


logger = logging.getLogger(__name__)


def normalise_kernel(matrix):
    return matrix / numpy.sqrt(numpy.outer(matrix.diagonal(), matrix.diagonal()))


def load_kernel_file(filename, normalise=True):
    kernel = numpy.load(filename)
    if normalise:
        return normalise_kernel(kernel)
    else:
        return kernel


def load_kernels(kernel_files, normalise=True):
    kernel_matrices = [load_kernel_file(x, normalise) for x in kernel_files]
    kernel_sum = numpy.sum(kernel_matrices, axis=0)
    if normalise:
        return normalise_kernel(kernel_sum)
    else:
        return kernel_sum / len(kernel_matrices)


# Hold the GNPS records
class GNPS:
    def __init__(self, filename):
        logger.debug(f"GNPS {sys._getframe().f_code.co_name}")
        self.data_gnps = scipy.io.loadmat(filename)
        self.data_fp_array = numpy.array(self.data_gnps["fp"].todense())

    def get(self, index):
        # maybe add output type conversions (from sparse mat.)?
        fingerprint = self.data_fp_array[:, [index]]
        inchi = self.data_gnps["inchi"][index][0][0]
        formula = self.data_gnps["mf"][index][0][0]

        return inchi, formula, fingerprint

    def get_fingerprints(self, indices):
        logger.debug(f"GNPS {sys._getframe().f_code.co_name}")
        return self.data_fp_array[:, indices].T

    def set_fingerprint(self, fingerprint):
        logger.debug(f"GNPS {sys._getframe().f_code.co_name}")
        fp_list = []
        num_fps = len(self.data_gnps["inchi"])
        count = 0
        for inchi in self.data_gnps["inchi"]:
            count += 1
            inchi = inchi[0][0]
            fp = calc_fp(inchi, fingerprint)
            fp_list.append(fp)
            if count % 100 == 0:
                logger.debug(f"Done {count} / {num_fps}")
        self.data_fp_array = numpy.array(fp_list).T

    def set_fingerprint_from_file(self, filename):
        logger.debug(f"GNPS {sys._getframe().f_code.co_name}")
        self.data_fp_array = numpy.load(filename)

    def save_fingerprint_to_file(self, filename):
        numpy.save(filename, self.data_fp_array)


def load_folds(filename):
    with open(filename) as f:
        fold_ids = f.readlines()
    fold_ids = [x.strip() for x in fold_ids]
    return fold_ids


def load_kernel(filename):
    with open(filename) as f:
        raw = f.read().strip().split()

    data = numpy.array([float(x.strip()) for x in raw])
    dim = int(numpy.sqrt(len(data)))
    data = data.reshape((dim, dim))
    return data


def load_avg_kernel(path):
    kernel_files = [
        "ALIGND.txt",
        "CP2Plus.txt",
        "CPI.txt",
        "CPK.txt",
        "LB.txt",
        "LI.txt",
        "NI.txt",
        "RLB.txt",
        "ALIGN.txt",
        "CP2.txt",
        "CPJB.txt",
        "CSC.txt",
        "LC.txt",
        "LW.txt",
        "NSF.txt",
        "RLI.txt",
        "CEC.txt",
        "CPC.txt",
        "CPJ.txt",
        "FIPP.txt",
        "LIPP.txt",
        "NB.txt",
        "PPKr.txt",
        "WPC.txt",
    ]

    kernel = 0
    for i in kernel_files:
        kernel += load_kernel(path + os.sep + i)
    return kernel / len(kernel_files)


def load_candidates(path):
    candidate_sets = {}
    for file in os.listdir(path):
        candidate_name = extract_candidate_name(file)
        candidate_sets[candidate_name] = []
        if not file.endswith("mat"):
            continue
        data = scipy.io.loadmat(path + os.sep + file)
        num_candidates, _ = data["inchi"].shape
        for i in range(num_candidates):
            cand_inchi = data["inchi"][i, 0][0]
            cand_fingerprint = data["fp"][:, [i]]
            candidate_sets[candidate_name].append((cand_inchi, cand_fingerprint))

    return candidate_sets


def load_candidate_file_fp(filename, fp_filename, fingerprint):
    candidates = []
    data = scipy.io.loadmat(filename)
    num_candidates, _ = data["inchi"].shape
    cand_inchi = data["inchi"][:, 0]
    if os.path.exists(fp_filename + ".npy"):
        fp_data = numpy.load(fp_filename + ".npy")
    else:
        fp_data = [calc_fp(x[0], fingerprint) for x in cand_inchi]
        numpy.save(fp_filename, fp_data)
    for i in range(num_candidates):
        inchi = cand_inchi[i][0]
        cand_fingerprint = fp_data[i]
        candidates.append((inchi, cand_fingerprint))
    return candidates


def calc_fp(inchi, fingerprint):
    fp = mk_fprints.fingerprint_from_inchi(inchi, fingerprint)
    return fp


def load_candidate_file(filename):
    candidates = []
    data = scipy.io.loadmat(filename)
    num_candidates, _ = data["inchi"].shape
    # fp_vectors = numpy.array(data['fp'].todense())
    fp_vectors = numpy.array(data["fp"].todense())
    for i in range(num_candidates):
        cand_inchi = data["inchi"][i, 0][0]
        # cand_fingerprint = data['fp'][:, i]
        cand_fingerprint = fp_vectors[:, i]
        candidates.append((cand_inchi, cand_fingerprint))
    return candidates


def extract_candidate_name(filename):
    return filename.split(".")[0].split("_")[-1]


def load_spectra(filename):
    spectra = []
    with open(filename) as f:
        for l in f.readlines():
            if l.startswith("SPECTRUM_ID"):
                continue
            spectrum_id, compound, inchi = l.strip().split("\t")
            spectra.append((spectrum_id, compound, inchi))
    return spectra


# Holds the data that the IOKR can request.
# Needs to be central so we can get the kernel values for new samples
# TODO:
# - Also handle output kernel values (queriable in the same way?)
# - more flexible way of loading (input) kernels
# - Calclulate novel input kernels
class IOKRDataServer:
    def __init__(self, path, kernel=None):
        logger.debug("IOKRDataServer {}".format(sys._getframe().f_code.co_name))
        self.path = path
        self.gnps = GNPS(os.path.join(path, "data_GNPS.mat"))
        # self.candidates = load_candidates(path + os.sep + 'candidates')
        self.spectra = load_spectra(path + os.sep + "spectra.txt")

        self.folds = numpy.array(load_folds(path + os.sep + "cv_ind.txt"))
        if kernel is None:
            logger.debug("No kernel specified. Please initialise manually.")
            self.kernel = None
        elif kernel == "avg":
            logger.debug("Loading average kernel.")
            self.kernel = load_avg_kernel(path + os.sep + "input_kernels")
        else:
            logger.debug("Loading kernel %s" % kernel)
            self.kernel = load_kernel(path + os.sep + "input_kernels" + os.sep + kernel)

        self.candidate_path = path + os.sep + "candidates" + os.sep + "candidate_set_%s.mat"

        self.dimension = None
        self.fingerprint = None
        self.ms = []

    def load_ms_files(self, path):
        logger.debug("IOKRDataServer {}".format(sys._getframe().f_code.co_name))
        if os.path.exists(os.path.join(path, "ms.npy")):
            logger.debug("Loading cached MS data")
            t = time.time()
            try:
                self.ms = numpy.load(open(os.path.join(path, "ms.npy"), "rb"), allow_pickle=True)
                logger.debug(f"Done in {time.time() - t:2f}s")
                return
            except Exception as e:
                logger.debug("Failed to load cached MS data ({}), regenerating it...".format(e))

        logger.debug("Loading .ms files")
        t = time.time()
        for spectrum_id in [x[0] for x in self.spectra]:
            ms = spectrum.MSSpectrum()
            ms.correct_for_ionisation = True
            ms.normalise = True
            ms.filter = spectrum_filters.filter_by_frozen_dag
            ms.load(os.path.join(path, "SPEC", spectrum_id + ".ms"))
            self.ms.append(ms)
        logger.debug(f"Loading took {time.time() - t:2f}s")
        logger.debug("Caching MS data")
        numpy.save(open(os.path.join(path, "ms.npy"), "wb"), self.ms)

    def get_candidates(self, formula):
        logger.debug("IOKRDataServer {}".format(sys._getframe().f_code.co_name))
        if self.fingerprint is None:
            yield from load_candidate_file(self.candidate_path % formula)
        else:
            candidate_path = self.candidate_path % formula
            fp_path = self.fp_path + os.sep + "fp_%s.bin" % formula
            yield from load_candidate_file_fp(candidate_path, fp_path, self.fingerprint)

    def get_sample(self, idx, skip_candidates=False):
        gnps_inchi, formula, fingerprint = self.gnps.get(idx)
        if skip_candidates:
            candidates = []
        else:
            candidates = self.get_candidates(formula)
        spectrum_id, spectrum_name, spectrum_inchi = self.spectra[idx]

        assert gnps_inchi == spectrum_inchi
        return {
            "inchi": gnps_inchi,
            "formula": formula,
            "fingerprint": fingerprint,
            "candidates": candidates,
            "spectrum_id": spectrum_id,
            "spectrum_name": spectrum_name,
        }

    def get_cv_set(self, label, complement=False):
        logger.debug("IOKRDataServer {}".format(sys._getframe().f_code.co_name))
        if label not in self.folds:
            return []
        if not complement:
            indices = numpy.where(self.folds == label)[0]
        else:
            indices = numpy.where(self.folds != label)[0]
        kernel_submatrix = self.kernel[numpy.ix_(indices, indices)]
        sample = [self.get_sample(x, skip_candidates=True) for x in indices]
        fp_matrix = numpy.hstack([x["fingerprint"] for x in sample]).T

        return kernel_submatrix, fp_matrix

    def get_latent_vector(self, index):
        return (self.get_sample(index)["fingerprint"]).T

    def kernel_product(self, index_1, index_2):
        return self.kernel[index_1, index_2]

    def kernel_product_set(self, index_1, indices):
        return self.kernel[index_1, indices]

    def get_kernel_matrix(self, indices):
        return self.kernel[numpy.ix_(indices, indices)]

    def get_latent_vectors(self, indices):
        sample = [self.get_sample(x, skip_candidates=True) for x in indices]
        return numpy.hstack([x["fingerprint"] for x in sample]).T

    def get_latent_vectors_vec(self, indices):
        fingerprints = self.gnps.get_fingerprints(indices)
        return fingerprints

    def get_dimension(self):
        if self.dimension is None:
            self.dimension = len(self.get_sample(0, skip_candidates=True)["fingerprint"])
        return self.dimension

    def get_indices(self, label, complement=False):
        if complement:
            indices = numpy.where(self.folds != label)[0]
        else:
            indices = numpy.where(self.folds == label)[0]
        return indices

    def get_all_indices(self):
        return [x for x in range(len(self.folds))]

    def set_fingerprint(self, fingerprint):
        logger.debug("IOKRDataServer {}".format(sys._getframe().f_code.co_name))
        self.fingerprint = fingerprint

        fppath = self.path
        fpfile = os.path.join(self.path, f"fp_{fingerprint}_gnps.bin.npy")
        # fppath = self.path + os.sep + 'fp_' + fingerprint + os.sep
        # fpfile = fppath + os.sep + 'fp_gnps.bin'
        self.fp_path = fppath
        logger.debug(f'fpfile is "{fpfile}"')
        if os.path.exists(fpfile):
            logger.debug("Loading GNPS fingerprints from file")
            self.gnps.set_fingerprint_from_file(fpfile)
        else:
            logger.debug("Recalculating GNPS fingerprints")
            self.gnps.set_fingerprint(fingerprint)
            logger.debug("Saving GNPS fingerprints to file")
            self.gnps.save_fingerprint_to_file(fpfile)

    def build_kernel_matrix(self):
        logger.debug("IOKRDataServer {}".format(sys._getframe().f_code.co_name))
        kernel_matrix = numpy.zeros((len(self.spectra), len(self.spectra)))
        for i in range(len(self.spectra)):
            for j in range(i):
                kernel_value = self.calculate_kernel(self.ms[i], self.ms[j])
                kernel_matrix[i, j] = kernel_value
                kernel_matrix[j, i] = kernel_value

        self.kernel = normalise_kernel(kernel_matrix)



from __future__ import annotations
import logging
from os import PathLike
from pyteomics import mgf
from nplinker.metabolomics import Spectrum
from nplinker.metabolomics.abc import SpectrumLoaderBase


logger = logging.getLogger(__name__)


class GNPSSpectrumLoader(SpectrumLoaderBase):
    """Load mass spectra from the given GNPS MGF file.

    ??? info "Concept"
        [GNPS data][gnps-data]

    The file mappings file is from GNPS output archive, as described below
    for each GNPS workflow type:

    1. METABOLOMICS-SNETS
        - METABOLOMICS-SNETS*.mgf
    2. METABOLOMICS-SNETS-V2
        - METABOLOMICS-SNETS-V2*.mgf
    3. FEATURE-BASED-MOLECULAR-NETWORKING
        - spectra/*.mgf
    """

    def __init__(self, file: str | PathLike) -> None:
        """Initialize the GNPSSpectrumLoader.

        Args:
            file: path to the MGF file.

        Raises:
            ValueError: Raises ValueError if the file is not valid.

        Examples:
            >>> loader = GNPSSpectrumLoader("gnps_spectra.mgf")
            >>> print(loader.spectra[0])
        """
        self._file = str(file)
        self._spectra: list[Spectrum] = []

        self._validate()
        self._load()

    @property
    def spectra(self) -> list[Spectrum]:
        """Get the list of Spectrum objects.

        Returns:
            list[Spectrum]: the loaded spectra as a list of `Spectrum` objects.
        """
        return self._spectra

    def _validate(self) -> None:
        """Validate GNPS MGF file.

        Raises:
            ValueError: Raises ValueError if the file is not valid.
        """
        # check the local scope of a single MS/MS query (spectrum) has the
        # required parameters. Note that this is not the header of the MGF
        # file, but the local scope of each spectrum.
        required_params = ["scans", "pepmass", "charge"]
        for spec in mgf.MGF(self._file):
            for param in required_params:
                if param not in spec["params"]:
                    raise ValueError(
                        f"Invalid MGF file '{self._file}'. "
                        f"Expected parameter '{param}' not found, "
                        f"but got '{spec['params']}'."
                    )

    def _load(self) -> None:
        """Load the MGF file into Spectrum objects."""
        for spec in mgf.MGF(self._file):
            # Skip if m/z array is empty, as this is an invalid spectrum.
            # The invalid spectrum does not exist in other GNPS files, e.g.
            # file mappings file and molecular families file. So we can safely
            # skip it.
            if len(spec["m/z array"]) == 0:
                continue

            # Load the spectrum
            spectrum_id: str = spec["params"]["scans"]
            # The pepmass in an mgf file is actually the m/z and not the peptide mass 
            # See: https://www.matrixscience.com/help/obsolete_data_file_formats.html
            precursor_mz: float = spec["params"]["pepmass"][0]
            precursor_charge: int = spec["params"]["charge"][0]
            rt = spec["params"].get("rtinseconds", 0)

            spectrum = Spectrum(
                id=spectrum_id,
                mz=list(spec["m/z array"]),
                intensity=list(spec["intensity array"]),
                precursor_mz=precursor_mz,
                precursor_charge=precursor_charge,
                rt=rt,
                metadata=spec["params"],
            )
            self._spectra.append(spectrum)



#!/usr/bin/env python
import sys
from setuptools import setup


if sys.version_info[:2] < (3, 9):
    raise RuntimeError("Python version >= 3.9 required.")

setup(scripts=["bin/install-nplinker-deps"])



from __future__ import annotations
from os import PathLike
from pathlib import Path
from typing_extensions import Self
from nplinker.utils import download_url
from .gnps_format import GNPSFormat
from .gnps_format import gnps_format_from_task_id


class GNPSDownloader:
    """Download GNPS zip archive for the given task id.

    ??? info "Concept"
        [GNPS data][gnps-data]

    Note that only GNPS workflows listed in the GNPSFormat enum are supported.

    Attributes:
        GNPS_DATA_DOWNLOAD_URL: URL template for downloading GNPS data.
        GNPS_DATA_DOWNLOAD_URL_FBMN: URL template for downloading GNPS data for FBMN.
        gnps_format: GNPS workflow type.
    """

    GNPS_DATA_DOWNLOAD_URL: str = (
        "https://gnps.ucsd.edu/ProteoSAFe/DownloadResult?task={}&view=download_clustered_spectra"
    )
    GNPS_DATA_DOWNLOAD_URL_FBMN: str = (
        "https://gnps.ucsd.edu/ProteoSAFe/DownloadResult?task={}&view=download_cytoscape_data"
    )

    def __init__(self, task_id: str, download_root: str | PathLike):
        """Initialize the GNPSDownloader.

        Args:
            task_id: GNPS task id, identifying the data to be downloaded.
            download_root: Path where to store the downloaded archive.

        Raises:
            ValueError: If the given task id does not correspond to a supported
                GNPS workflow.

        Examples:
            >>> GNPSDownloader("c22f44b14a3d450eb836d607cb9521bb", "~/downloads")
        """
        gnps_format = gnps_format_from_task_id(task_id)
        if gnps_format == GNPSFormat.Unknown:
            raise ValueError(
                f"Unknown workflow type for GNPS task '{task_id}'."
                f"Supported GNPS workflows are described in the GNPSFormat enum, "
                f"including such as 'METABOLOMICS-SNETS', 'METABOLOMICS-SNETS-V2' "
                f"and 'FEATURE-BASED-MOLECULAR-NETWORKING'."
            )

        self._task_id = task_id
        self._download_root: Path = Path(download_root)
        self._gnps_format = gnps_format
        self._file_name = gnps_format.value + "-" + self._task_id + ".zip"

    @property
    def gnps_format(self) -> GNPSFormat:
        """Get the GNPS workflow type.

        Returns:
            GNPS workflow type.
        """
        return self._gnps_format

    def download(self) -> Self:
        """Download GNPS data.

        Note: GNPS data is downloaded using the POST method (empty payload is OK).
        """
        download_url(
            self.get_url(), self._download_root, filename=self._file_name, http_method="POST"
        )
        return self

    def get_download_file(self) -> str:
        """Get the path to the downloaded file.

        Returns:
            Download path as string
        """
        return str(Path(self._download_root) / self._file_name)

    def get_task_id(self) -> str:
        """Get the GNPS task id.

        Returns:
            Task id as string.
        """
        return self._task_id

    def get_url(self) -> str:
        """Get the download URL.

        Returns:
            URL pointing to the GNPS data to be downloaded.
        """
        if self.gnps_format == GNPSFormat.FBMN:
            return GNPSDownloader.GNPS_DATA_DOWNLOAD_URL_FBMN.format(self._task_id)
        return GNPSDownloader.GNPS_DATA_DOWNLOAD_URL.format(self._task_id)



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy


class InputOutputKernelRegression:
    def __init__(self, data):
        self.data = data
        self.kernel_vector_cache = {}

    def set_training_indices(self, indices=None, _lambda=0.1):
        self._lambda = _lambda
        if indices is None:
            indices = range(self.data.data_size)
        self.training_set = indices

    def fit(self):
        training_data_kernel = self.data.get_kernel_matrix(self.training_set)
        training_data_latent = self.data.get_latent_vectors(self.training_set)

        eye = numpy.eye(len(training_data_kernel))

        training_data_latent = numpy.array(training_data_latent).T

        latent_basis = numpy.linalg.inv(self._lambda * eye + training_data_kernel)
        self.latent_basis = latent_basis
        self.basis = numpy.dot(training_data_latent, latent_basis)

    def calculate_fingerprint_kernel_vector(self, fingerprint, kernel="gaussian"):
        if kernel == "gaussian":

            def k(a, b, gamma=0.01):
                # kernel function
                d_sq = numpy.sum(numpy.power((a - b), 2))
                return numpy.exp(-gamma * d_sq)

            def k_vec(a_mat, b, gamma=0.01):
                # vectorised kernel function
                d_sq = numpy.sum(numpy.power((a_mat - b), 2), axis=1)
                return numpy.exp(-gamma * d_sq)

        # fp_id = hash(str(fingerprint))
        # if fp_id in self.kernel_vector_cache:
        #    return self.kernel_vector_cache[fp_id]

        training_data_latent = self.data.get_latent_vectors_vec(self.training_set)
        kernel_vector = k_vec(training_data_latent, fingerprint.T)
        # self.kernel_vector_cache[fp_id] = kernel_vector
        return kernel_vector

    def project_candidate(self, index, fingerprint):
        fingerprint_kernel_vector = self.calculate_fingerprint_kernel_vector(fingerprint)
        x_kernel_vector = self.data.kernel_product_set(index, self.training_set)
        res = numpy.dot(numpy.dot(fingerprint_kernel_vector, self.latent_basis), x_kernel_vector)
        return res

    def rank_candidates(self, index, candidate_fingerprints):
        candidate_distances = [
            self.project_candidate(index, fingerprint) for fingerprint in candidate_fingerprints
        ]
        return [
            x[1]
            for x in sorted(
                zip(candidate_distances, range(len(candidate_distances))),
                key=lambda x: x[0],
                reverse=True,
            )
        ]

    def project(self, index):
        x_kernel_vector = self.data.kernel_product_set(index, self.training_set)
        projection = numpy.dot(self.basis, x_kernel_vector)
        return projection

    def test(self, index, cutoff=0.01):
        proj = self.project(index)
        return [1 if x > cutoff else 0 for x in proj]


def main():
    target_vectors = numpy.array(
        [[0, 1, 0], [1, 0, 0], [1, 1, 0], [1, 1, 1], [0, 0, 1], [0, 1, 1], [1, 0, 1]], dtype="float"
    )
    repr_vectors = numpy.array(
        [[0, 1, 0], [1, 0, 0], [1, 1, 0], [1, 1, 1], [0, 0, 1], [0, 1, 1], [1, 0, 1]], dtype="float"
    )

    kernel_matrix = numpy.zeros((7, 7))
    for i in range(7):
        for j in range(i + 1):
            kernel_matrix[i, j] = kernel_matrix[j, i] = numpy.dot(repr_vectors[i], repr_vectors[j])

    data = DataStore(kernel_matrix, target_vectors)
    okr = InputOutputKernelRegression(data)
    okr.set_training_indices([0, 2, 3], _lambda=0.0)
    okr.fit()
    print(okr.test(0), target_vectors[0])
    print(okr.test(1), target_vectors[1])
    print(okr.test(2), target_vectors[2])
    print(okr.test(3), target_vectors[3])


if __name__ == "__main__":
    main()



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math


def fast_cosine_shift(spectrum1, spectrum2, tol, min_match):
    if len(spectrum1.peaks) == 0 or len(spectrum2.peaks) == 0:
        return 0.0, []

    spec1 = sqrt_normalise(spectrum1.peaks)
    spec2 = sqrt_normalise(spectrum2.peaks)

    zero_pairs = find_pairs(spec1, spec2, tol, shift=0.0)

    shift = spectrum1.precursor_mz - spectrum2.precursor_mz

    nonzero_pairs = find_pairs(spec1, spec2, tol, shift=shift)

    matching_pairs = zero_pairs + nonzero_pairs

    matching_pairs = sorted(matching_pairs, key=lambda x: x[2], reverse=True)

    used1 = set()
    used2 = set()
    score = 0.0
    used_matches = []
    for m in matching_pairs:
        if m[0] not in used1 and m[1] not in used2:
            score += m[2]
            used1.add(m[0])
            used2.add(m[1])
            used_matches.append(m)
    if len(used_matches) < min_match:
        score = 0.0
    return score, used_matches


def find_pairs(spec1, spec2, tol, shift=0):
    matching_pairs = []
    spec2lowpos = 0
    spec2length = len(spec2)

    for idx, (mz, intensity) in enumerate(spec1):
        # do we need to increase the lower idx?
        while spec2lowpos < spec2length and spec2[spec2lowpos][0] + shift < mz - tol:
            spec2lowpos += 1
        if spec2lowpos == spec2length:
            break
        spec2pos = spec2lowpos
        while spec2pos < spec2length and spec2[spec2pos][0] + shift < mz + tol:
            matching_pairs.append((idx, spec2pos, intensity * spec2[spec2pos][1]))
            spec2pos += 1

    return matching_pairs


def fast_cosine(spectrum1, spectrum2, tol, min_match):
    # spec 1 and spec 2 have to be sorted by mz
    if len(spectrum1.peaks) == 0 or len(spectrum2.peaks) == 0:
        return 0.0, []
    # find all the matching pairs

    spec1 = sqrt_normalise(spectrum1.peaks)
    spec2 = sqrt_normalise(spectrum2.peaks)

    matching_pairs = find_pairs(spec1, spec2, tol, shift=0.0)

    matching_pairs = sorted(matching_pairs, key=lambda x: x[2], reverse=True)
    used1 = set()
    used2 = set()
    score = 0.0
    used_matches = []
    for m in matching_pairs:
        if m[0] not in used1 and m[1] not in used2:
            score += m[2]
            used1.add(m[0])
            used2.add(m[1])
            used_matches.append(m)
    if len(used_matches) < min_match:
        score = 0.0
    return score, used_matches


def comp_scores(spectra, file_scan, similarity_function, similarity_tolerance, min_match):
    # a method for testing -- just computes scores between a bunch of scans
    specs = []
    for file_name, scan_number in file_scan:
        specs.append(
            filter(lambda x: x.file_name == file_name and x.scan_number == scan_number, spectra)[0]
        )

    for i in range(len(file_scan) - 1):
        for j in range(i + 1, len(file_scan)):
            (f, s) = file_scan[i]
            spec = specs[i]
            (f2, s2) = file_scan[j]
            spec2 = specs[j]
            sc, _ = similarity_function(spec, spec2, similarity_tolerance, min_match)
            print(f"{f},{s} <-> {f2},{s2} = {sc}")


def sqrt_normalise(peaks):
    temp = []
    total = 0.0
    for mz, intensity in peaks:
        temp.append((mz, math.sqrt(intensity)))
        total += intensity
    norm_facc = math.sqrt(total)
    normalised_peaks = []
    for mz, intensity in temp:
        normalised_peaks.append((mz, intensity / norm_facc))
    return normalised_peaks



from __future__ import annotations
import re
import zipfile
from enum import Enum
from enum import unique
from os import PathLike
from pathlib import Path
import httpx
from bs4 import BeautifulSoup


GNPS_TASK_URL = "https://gnps.ucsd.edu/ProteoSAFe/status.jsp?task={}"


@unique
class GNPSFormat(Enum):
    """Enum class for GNPS formats or workflows.

    ??? info "Concept"
        [GNPS data][gnps-data]

    The name of the enum is a short name for the workflow, and the value of the enum is the workflow
    name used on the GNPS website.
    """

    # Format: ShortName = "GNPSWorkflowName"
    SNETS = "METABOLOMICS-SNETS"
    SNETSV2 = "METABOLOMICS-SNETS-V2"
    FBMN = "FEATURE-BASED-MOLECULAR-NETWORKING"
    Unknown = "Unknown-GNPS-Workflow"


def gnps_format_from_task_id(task_id: str) -> GNPSFormat:
    """Detect GNPS format for the given task id.

    Args:
        task_id: GNPS task id.

    Returns:
        The format identified in the GNPS task.

    Examples:
        >>> gnps_format_from_task_id("c22f44b14a3d450eb836d607cb9521bb")
        <GNPSFormat.SNETS: 'METABOLOMICS-SNETS'>
        >>> gnps_format_from_task_id("189e8bf16af145758b0a900f1c44ff4a")
        <GNPSFormat.SNETSV2: 'METABOLOMICS-SNETS-V2'>
        >>> gnps_format_from_task_id("92036537c21b44c29e509291e53f6382")
        <GNPSFormat.FBMN: 'FEATURE-BASED-MOLECULAR-NETWORKING'>
        >>> gnps_format_from_task_id("0ad6535e34d449788f297e712f43068a")
        <GNPSFormat.Unknown: 'Unknown-GNPS-Workflow'>
    """
    task_html = httpx.get(GNPS_TASK_URL.format(task_id))
    soup = BeautifulSoup(task_html.text, features="html.parser")
    try:
        # find the td tag that follows the th tag containing 'Workflow'
        workflow_tag = soup.find("th", string="Workflow").find_next_sibling("td")  # type: ignore
        workflow_format = workflow_tag.contents[0].strip()  # type: ignore
    except AttributeError:
        return GNPSFormat.Unknown

    if workflow_format == GNPSFormat.FBMN.value:
        return GNPSFormat.FBMN
    if workflow_format == GNPSFormat.SNETSV2.value:
        return GNPSFormat.SNETSV2
    if workflow_format == GNPSFormat.SNETS.value:
        return GNPSFormat.SNETS
    return GNPSFormat.Unknown


def gnps_format_from_archive(zip_file: str | PathLike) -> GNPSFormat:
    """Detect GNPS format from GNPS zip archive.

    The detection is based on the filename of the zip file and the names of the
    files contained in the zip file.

    Args:
        zip_file: Path to the GNPS zip file.

    Returns:
        The format identified in the GNPS zip file.

    Examples:
        >>> gnps_format_from_archive("ProteoSAFe-METABOLOMICS-SNETS-c22f44b1-download_clustered_spectra.zip")
        <GNPSFormat.SNETS: 'METABOLOMICS-SNETS'>
        >>> gnps_format_from_archive("ProteoSAFe-METABOLOMICS-SNETS-V2-189e8bf1-download_clustered_spectra.zip")
        <GNPSFormat.SNETSV2: 'METABOLOMICS-SNETS-V2'>
        >>> gnps_format_from_archive("ProteoSAFe-FEATURE-BASED-MOLECULAR-NETWORKING-672d0a53-download_cytoscape_data.zip")
        <GNPSFormat.FBMN: 'FEATURE-BASED-MOLECULAR-NETWORKING'>
    """
    file = Path(zip_file)
    # Guess the format from the filename of the zip file
    if GNPSFormat.FBMN.value in file.name:
        return GNPSFormat.FBMN
    # the order of the if statements matters for the following two
    if GNPSFormat.SNETSV2.value in file.name:
        return GNPSFormat.SNETSV2
    if GNPSFormat.SNETS.value in file.name:
        return GNPSFormat.SNETS

    # Guess the format from the names of the files in the zip file
    with zipfile.ZipFile(file) as archive:
        filenames = archive.namelist()
    if any(GNPSFormat.FBMN.value in x for x in filenames):
        return GNPSFormat.FBMN
    # the order of the if statements matters for the following two
    if any(GNPSFormat.SNETSV2.value in x for x in filenames):
        return GNPSFormat.SNETSV2
    if any(GNPSFormat.SNETS.value in x for x in filenames):
        return GNPSFormat.SNETS

    return GNPSFormat.Unknown


def gnps_format_from_file_mapping(file: str | PathLike) -> GNPSFormat:
    """Detect GNPS format from the given file mapping file.

    The GNPS file mapping file is located in different folders depending on the
    GNPS workflow. Here are the locations in corresponding GNPS zip archives:

    - `METABOLOMICS-SNETS` workflow: the `.tsv` file in the folder
        `clusterinfosummarygroup_attributes_withIDs_withcomponentID`
    - `METABOLOMICS-SNETS-V2` workflow: the `.clustersummary` file (tsv) in the folder
        `clusterinfosummarygroup_attributes_withIDs_withcomponentID`
    - `FEATURE-BASED-MOLECULAR-NETWORKING` workflow: the `.csv` file in the folder
        `quantification_table`

    Args:
        file: Path to the file to peek the format for.

    Returns:
        GNPS format identified in the file.
    """
    with open(file, "r") as f:
        header = f.readline().strip()

    if re.search(r"\bAllFiles\b", header):
        return GNPSFormat.SNETS
    if re.search(r"\bUniqueFileSources\b", header):
        return GNPSFormat.SNETSV2
    if re.search(r"\b{}\b".format(re.escape("row ID")), header):
        return GNPSFormat.FBMN
    return GNPSFormat.Unknown



from __future__ import annotations
from dataclasses import dataclass
from dataclasses import fields
from .scoring_method import ScoringMethod


@dataclass
class Score:
    """A data class to represent score data.

    Attributes:
        name: the name of the scoring method. See `ScoringMethod` for valid values.
        value: the score value.
        parameter: the parameters used for the scoring method.
    """

    name: str
    value: float
    parameter: dict

    def __post_init__(self) -> None:
        """Check if the value of `name` is valid.

        Raises:
            ValueError: if the value of `name` is not valid.
        """
        if ScoringMethod.has_value(self.name) is False:
            raise ValueError(
                f"{self.name} is not a valid value. Valid values are: {[e.value for e in ScoringMethod]}"
            )

    def __getitem__(self, key):
        if key in {field.name for field in fields(self)}:
            return getattr(self, key)
        else:
            raise KeyError(f"{key} not found in {self.__class__.__name__}")

    def __setitem__(self, key, value):
        # validate the value of `name`
        if key == "name" and ScoringMethod.has_value(value) is False:
            raise ValueError(
                f"{value} is not a valid value. Valid values are: {[e.value for e in ScoringMethod]}"
            )

        if key in {field.name for field in fields(self)}:
            setattr(self, key, value)
        else:
            raise KeyError(f"{key} not found in {self.__class__.__name__}")



# Copyright 2021 The NPLinker Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import subprocess
import sys


logger = logging.getLogger(__name__)


def run_canopus(mgf_file, output_path, extra_params="--maxmz 600 formula zodiac structure canopus"):
    """Runs canopus from the sirius workflow.

    Args:
        mgf_file: str, path to the mgf file with spectra info
        output_path: str, path to canopus_dir for results
        extra_params: str, the extra parameters for running canopus. this
            should always end with canopus, default:
            --maxmz 600 formula zodiac structure canopus
    Returns:
        True if everything runs okay

    Within NPLinker a normal (default) will look like:
    sirius -i METABOLOMICS-SNETS-V2-<id>-download_clustered_spectra-main.mgf
        -o canopus --maxmz 600 formula zodiac structure canopus

    Note that SIRIUS might take a long time when setting a high --maxmz cutoff
    """
    logger.info(
        'run_canopus: input="{}", output="{}", extra_params={}"'.format(
            mgf_file, output_path, extra_params
        )
    )
    if "canopus" not in extra_params and " C " not in extra_params:
        logger.warning("canopus not in sirius command, please check canopus parameters")

    if os.path.exists(os.path.join(output_path, "completed")):
        logger.info("CANOPUS appears to have been run already, skipping!")
        logger.info("To force re-run, delete {}".format(os.path.join(output_path, "completed")))
        return True

    try:
        subprocess.run(["sirius", "-h"], capture_output=True)
    except Exception as e:
        raise Exception(f"Failed to find/run SIRIUS (CANOPUS) (err={e})")

    if not os.path.exists(mgf_file):
        raise Exception(f'mgf input file "{mgf_file}" does not exist!')

    # configure the IO-related parameters
    args = ["sirius", "-i", mgf_file, "-o", output_path]
    if "--log" not in args:
        args.extend(["--log", "INFO"])

    # append the user supplied params, if any
    if len(extra_params) > 0:
        args.extend(extra_params.split(" "))

    logger.info(f"CANOPUS command: {args}")
    result = subprocess.run(args, stdout=sys.stdout, stderr=sys.stderr)
    logger.info("CANOPUS completed with return code {}".format(result.returncode))
    # use subprocess.CompletedProcess.check_returncode() to test if the CANOPUS
    # process exited successfully. This throws an exception for non-zero returncodes
    # which will indicate to the PODPDownloader module that something went wrong.
    result.check_returncode()

    # use presence of this file as a quick way to check if a previous run
    # finished or not
    open(os.path.join(output_path, "completed"), "w").close()

    return True
